/* GENEHUNTER - Kruglyak, Daly, Reeve-Daly, and Lander
   Whitehead Institute for Biomedical Research, Copyright 1996
   modified by Konstantin Strauch, Johannes Dietter, Manuel Mattheisen,
   and Markus Brugger.    Copyright (c) 1999-2019 */

#include <search.h>
#define INC_LIB
#define INC_SHELL
#include "npl.h"

#include <stdio.h>
#include <stdlib.h>
#include <assert.h>
#include <math.h>

#define real_eq(x,y) (fabs((x)-(y))<1e-10)
#define MAX(X, Y) ((X) < (Y) ? (Y) : (X))
#define MIN(X, Y) ((X) < (Y) ? (X) : (Y))

/* externals */
LOCUS_INFO *locus;
PHENOTYPE_INFO *phenotype;
int num_markers, num_in_map_order, *map_order;
int num_phenotypes, num_covs; /*** new ***/
int num_loci; /*** new ***/ /* num_markers + num_phenotypes + num_covs*/
int ind_sp;
int nr_marker; /*number of markers read from map*/
int nr_omitted;/*number of omitted markers in map*/
double *map_distance;
double *map_distance_m;
double *map_distance_f;
MARKER *supp_marker;
MARKER *first_marker;
MARKER *last_marker;
int nr_miss_marker;
int ind_use_map_marker;
double affected_freq;
int num_liability_classes;
double penetrance[MAX_LIABILITY_CLASSES+1][6];
double apriori_disease[MAX_LIABILITY_CLASSES+1][6][6];
int recode_liab_class[MAX_LIABILITY_CLASSES+1]; /* for rec. of liab.classes */
int old_num_liability_classes;                  /* new 04/2004 (K. Strauch) */

static int phenotypes_read, covs_read;  /* added (K. Strauch 04/2008) */

double save_affected_freq;
double save_penetrance[MAX_LIABILITY_CLASSES+1][6];

/* internals */

int miss_marker[MAXLOCI];

void read_linkage_marker_file(FILE*); /* arg: fp 
				         Loop over LINKAGE loci file */
void read_linkage_marker(FILE*,int);  /* args: fp, loci_read 
				         Read indiv LINKAGE marker entry */
void free_map_structs(void);          /* only called if load_markers
					 cannot completely load the 
					 file */

/* Associated procedures in newcmds.c:
   void print_map_order(void) -- print out the current map
   int lookup_locus(char*) -- find the index of a locus in global locus array
   command set_use_map(void) -- input marker distances
*/
void get_map_marker_decode(char* map_name, char* map_file, int* nr_marker, int *nr_omitted);
void get_map_marker_marshfield(char* map_name, char* map_file, int* nr_marker, int *nr_omitted);
void get_map_marker_duffy(char* map_name, char* map_file, int* nr_marker, int *nr_omitted);
void get_map_marker_rutgers(char* map_name, char* map_file, int* nr_marker, int *nr_omitted);
void get_map_marker_nievergelt(char* map_name, char* map_file, int* nr_marker, int* nr_omitted);
void get_map_marker(char* map_name, char* map_file, int* nr_marker, int *nr_omitted);
void fill_marker_structs(char* map_name, int* nr_marker);
void check_user_marker_build_distances(void);
void check_map_marker(struct marker *start);
void repair_map_marker_positions(struct marker* start);
void dls_store(struct marker *i, struct marker **start, struct marker **last);

command load_markers(void)
{

  /* The upper level command.  Reads the number of markers
     in the file, determines the type of file and then
     passes control off to the appropriate file to actually handle
     the input. */

  char *filename;
  int i, num_fields,foo1,foo2,foo3;
  FILE *fp;


  /* Check that the locus file hasn't already been loaded */
  if (num_markers > 0) {
    sf(ps,"Genetic marker info is already loaded for %d markers.\n",
       num_markers); pr();
    print("This command will therefore be ignored.\n");
    return;
  }

  first_marker = NULL;  last_marker= NULL;
  phenotypes_read = FALSE;  covs_read = FALSE;
  /* added (K. Strauch 04/2008) */

  fp=NULL; locus=NULL;
  run {
    /* If no filename is given, just print the status */
    filename = get_temp_string();
    use_uncrunched_args();
    get_one_arg(stoken,"",filename);
    
    if (nullstr(filename)) {
      /* Print the status of the data */
      print("no data is currently loaded - use 'load markers <filename>'\n");
    }
    else {
      fp=open_file(filename,READ);
      
      /* Read the first line -  num_loci is all we use, risk
	 locus, sexlinked status, etc if present (in Linkage
	 style files) are discarded */
      _fgetln(fp);
      num_fields = sscanf(ln,"%d %d %d %d",&num_loci,&foo1,&foo2,&foo3);
      if (num_fields == 0) {
	sf(ps,"missing num_loci on first line of '%s'",filename);
	error(ps);
      }
      /* allocate locus array now that we know how many there are,
	 this array will be filled in as we read the rest of the file */
      array(locus, num_loci, LOCUS_INFO);          /*** This is a little screwy, because each of these will hold only ***/
      array(phenotype, num_loci, PHENOTYPE_INFO);  /***	the marker data, or the phenotype data, not both. ***/
      
      /* Number of fields on the first line gives us the type of
	 file.  Our own internal style just has the number of
	 markers on the first line. */
      if (num_fields == 4) {
	/* Linkage style data file */
	num_loci--; /* Linkage files count the disease locus */
	print("Parsing Linkage marker data file...\n");
	frewind(fp);
	read_linkage_marker_file(fp);
      }
      else {
	/* Parse our own internal format */
	error("Unrecognized marker file format");
      }
      close_file(fp); fp=NULL;

      /* skip the following if read_linkage_marker_file() did not finish */

      if (num_markers > 0) {
      /* Give the user some output info */
	sf(ps,"\n%d markers read from Linkage marker data file (last one = %s)\n", num_markers,
	   locus[num_markers-1].name); pr();

	if (num_in_map_order > 0) print("The selected markers and their positions can be found in the file\n'user_markers_in_map'.\n");
	else print("No marker set has been selected for the analysis yet.\nPlease specify markers and distances with the 'use' command.\n");

	if (modcalc && (modliab > num_liability_classes)) {
	  sf(ps,"\nWarning: liability class to be optimized (%d) does not exist.\n", modliab);
	  pr();
	  print("Use the 'liability class' command in order to change this.\n");
	}

	/* the following is for possible recoding of liability classes
	   by the command 'join liability classes' later (in newcmds.c) */
	old_num_liability_classes = num_liability_classes;
	for (i=0; i<=num_liability_classes; i++) recode_liab_class[i] = i;
      }
    }
  } m_except {
    when ENDOFILE: 
       print("Unexpected end-of-file, no markers read\n");
       close_file(fp); fp=NULL;
       free_map_structs();
       break;
    when CANTOPEN:
       sf(ps,"Can't open file '%s'\n",filename); 
       pr(); break;
    default:
       if (fp != NULL) { close_file(fp); fp=NULL;}
       free_map_structs(); 
       relay_messages;
  }
}

int count_digits(int n)
{
    int digits = 0;
    while (n > 0) {
        ++digits;
        n /= 10;
    }
    return digits;
}

int compute_max_line_length(int max_loci)
{
    assert(max_loci > 0);
    /* We need (max_loci - 1) spaces between the numbers.  Further,
       there are at most 9 1-digit numbers, 90 2-digit numbers, 900
       3-digit numbers, and so on.  We just sum up the resulting
       number of digits, add the number of spaces and 1 more digit for
       the final nul byte. */
    int d = count_digits(max_loci);  /* digits in max_loci */
    int n;  /* number of characters needed for "1 2 3 ... max_loci" */
    n = max_loci;  /* (max_loci - 1) spaces + \0 */
    while (d > 0) {
        if (d == 1) {
            n += MIN(max_loci, 9);
        } else {
            n += d * MAX(0, MIN(9 * pow(10, d-1), max_loci - 9 * pow(10, d-2)));
        }
        --d;
    }
    return n;
}

void read_linkage_marker_file(FILE *fp)
{

  /* This method was ported from HOMOZ to read in the locus
     file in LINKAGE format.  Some of the infomation read in
     is not actually used but it is left here to check the format
     of the file and in case it is needed at a later time. */

  int i,num1,num2,loci_read=0,this_loc,*new_order,count;
  volatile int cM_dists;
  char *strptr,*map_ord_str, *map_ord_str_cp;
  int *map_order_chk, num_in_map_order_chk;  
  double rnum1, rnum2, rnum3, rnum4, dist;
  void init_disease_info(void);
  int max_line_length = compute_max_line_length(MAXLOCI);

  run {

    _fgetln(fp);			/* header line, num_loci already read */
    _fgetln(fp);			/* mutation line - skip*/
    _fgetln(fp);                 /* map order */
    array(strptr,max_line_length,char); strcpy(strptr,ln); map_ord_str=strptr; map_ord_str_cp=strptr;
    _fgetln(fp);			/* should be the start of affectation locus */

    /*** LINKAGE formatting checks - might want to take this out ***/
    /* is this an acceptable affectation locus? */
    if (sscanf(ln,"%d %d",&num1,&num2)!=2) {
      sf(ps,"bad affectation (4th) line '%s'",ln);
      error(ps);
    }
    _fgetln(fp);			/* Read in the line with gene frequencies */
    if (sscanf(ln,"%lf %lf",&rnum1,&rnum2)!=2) {
      sf(ps,"bad disease gene frequency line (%s) - needs 2 numbers",ln);
      error(ps);
    }
    if (rnum1+rnum2 < 0.999 || rnum1+rnum2 > 1.001) {
      sf(ps,"disease allele frequencies (%.4lf, %.4lf) do not sum to 1.0",
	 rnum1,rnum2); error(ps);
    }
    save_affected_freq = affected_freq = rnum2;

    _fgetln(fp);			/* liability class line */
    if (!itoken(&ln,iREQUIRED,&num1)) {
      sf(ps,"bad liability class line '%s'",ln); error(ps);
    }
    num_liability_classes=num1;
    if (num_liability_classes > MAX_LIABILITY_CLASSES) {
      sf(ps,"too many liability classes (%d) - edit npl.h",num_liability_classes);
      error(ps);
    }
    /* get and store a line of penetrance data for each liability class... */
    for (i=0; i<num_liability_classes; i++) {
      _fgetln(fp);
      if (!imprinting) {
        if ((count=sscanf(ln,"%lf %lf %lf %lf",
			  &rnum1,&rnum2,&rnum3,&rnum4))==4) {
	  sf(ps,"bad penetrance line (%s) - needs only 3 numbers (imprinting is off)",ln);
	  error(ps);
	} else if (count != 3) {
          sf(ps,"bad penetrance line (%s) - needs 3 numbers",ln);
          error(ps);
        }
	save_penetrance[i+1][HOM_UNAFF] = penetrance[i+1][HOM_UNAFF] = rnum1;
	save_penetrance[i+1][HET_AU] = penetrance[i+1][HET_AU] = rnum2;
	save_penetrance[i+1][HET_UA] = penetrance[i+1][HET_UA] = rnum2;
	save_penetrance[i+1][HOM_AFF] = penetrance[i+1][HOM_AFF] = rnum3;
      } else {
        if (sscanf(ln,"%lf %lf %lf %lf",&rnum1,&rnum2,&rnum3,&rnum4)!=4) {
          sf(ps,"bad penetrance line (%s) - needs 4 numbers (imprinting is on)",ln);
          error(ps);
        }
	save_penetrance[i+1][HOM_UNAFF] = penetrance[i+1][HOM_UNAFF] = rnum1;
	save_penetrance[i+1][HET_AU] = penetrance[i+1][HET_AU] = rnum2;
	save_penetrance[i+1][HET_UA] = penetrance[i+1][HET_UA] = rnum3;
	save_penetrance[i+1][HOM_AFF] = penetrance[i+1][HOM_AFF] = rnum4;
      }
    }
    /* store some disease probabilities */
    init_disease_info();

    /*** End of LINKAGE formatting check, time to actually read the
      marker information ***/

    /**** Read in marker and phenotype information ****/
    for (i=0; i<num_loci; i++) {
      read_linkage_marker(fp,i);
      loci_read++;
    }     
    if (num_markers == 0) error("no markers found in linkage marker file");
            
    _fgetln(fp); while(nullstr(ln)) _fgetln(fp); /* gets sex diff. line */
    
    if (!itoken(&ln,iREQUIRED,&ind_sp)){ 
      error("Please indicate with the usual linkage notation whether you want to use\nsex-specific or sex-averaged recombination fractions.\n0: sex-averaged recombination fractions (only one line with cM or recfracs)\n2: sex-specific recombination fractions (one line with male cM or recfracs\n   followed by a second line with female cM or recfracs)");
    }else{
      if (ind_sp==2){
        ind_sp=1;
        sf(ps,"sex specific recombination frequencies are being used");pr();nl();
      }else if(ind_sp==0){
        sf(ps,"sex averaged recombination frequencies are being used");pr();nl();
      }else{
        error("Please indicate with the usual linkage notation whether you want to use\nsex-specific or sex-averaged recombination fractions.\n0: sex-averaged recombination fractions (only one line with cM or recfracs)\n2: sex-specific recombination fractions (one line with male cM or recfracs\n   followed by a second line with female cM or recfracs)");
      }  
    }    
    /* get male recombination frequencies */

    if (ind_use_map_marker==0) _fgetln(fp); /* list of recombination distances */
    
    /* fill globals with linkage map info */

    num_in_map_order=0;

    array(map_order, num_markers+1, int);
    array(map_distance, num_markers+1, double);
    array(map_distance_m, num_markers+1, double);
    array(new_order, num_markers+1, int);

    if (num_markers == 1) {
        num_in_map_order=1;
	map_order[0]=0;
    } else {
        /* read in recombination distance from end of file */
        count=0;
        while (itoken(&map_ord_str,iREQUIRED,&this_loc)) {
	  new_order[count] = this_loc;
	  count++;
	}
	if (new_order[0] == 1) {
	  cM_dists=FALSE;
	  /* disease listed first, skip i=0, first distance */
	  if (ind_use_map_marker==0) 
	    if (!rtoken(&ln,rREQUIRED,&dist)) 
	      error("insufficient number of recombination distances");
	  for (i=1; i<count-1; i++) {
	    map_order[i-1] = new_order[i]-2;
	    if (ind_use_map_marker==0){
	      if (!rtoken(&ln,rREQUIRED,&dist)) 
	        error("insufficient number of recombination distances");
	      if (dist < 0.0)
		error("recombination distance < 0");
	      map_distance_m[i-1] = dist;
	      if (dist >= 0.50) cM_dists=TRUE;
	    }  
	  }
	  if (count >= 2) {  /* otherwise no marker specified on line 3 */
	    map_order[count-2] = new_order[count-1]-2;
	    num_in_map_order=count-1;
	  }
	  if (ind_use_map_marker==0 && rtoken(&ln,rREQUIRED,&dist)) {
	      print("Warning: found more map distances than needed");
	      if (ind_sp == 1) print(" (male map)\n"); else nl();
	  }
	  /*** rec_to_dist will be done later
	  if (ind_use_map_marker==0)
	    if (!cM_dists) {
	      for (i=0; i<count-2; i++) 
	        map_distance_m[i] = rec_to_dist(map_distance_m[i]);
	    }
	  ***/

	} else {

	  for (i=0; i<count; i++) {
	    if (new_order[i]==1) error("disease locus cannot be in map");
	    map_order[i]=new_order[i]-2;
	  }
	  cM_dists=FALSE;
	  if (ind_use_map_marker==0){
	    for (i=0; i<count-1; i++) {
	      /* do a better job of conversion here */
	      if (!rtoken(&ln,rREQUIRED,&dist)) 
	        error("insufficient number of recombination distances");
	      if (dist < 0.0)
		error("recombination distance < 0");
	      map_distance_m[i] = dist;
	      if (dist >= 0.50) cM_dists=TRUE;
	    }
	    if (rtoken(&ln,rREQUIRED,&dist)) {
	      print("Warning: found more map distances than needed");
	      if (ind_sp == 1) print(" (male map)\n"); else nl();
	    }
	    /*** rec_to_dist will be done later
	    if (!cM_dists) {
	      for (i=0; i<count-1; i++) 
	        map_distance_m[i] = rec_to_dist(map_distance_m[i]);
	    }
	    ***/
	  }  
	  num_in_map_order=count;
	}
    }
    unarray(new_order, int);
    /* unarray(strptr, char); */
  

    /* get female recombination frequencies */
    
    /* fill globals with linkage map info */
    num_in_map_order_chk=0;

    array(map_order_chk, num_markers+1, int);
    array(map_distance_f, num_markers+1, double);
    array(new_order, num_markers+1, int);

    if (num_markers == 1) {
       num_in_map_order_chk=1;
       map_order_chk[0]=0;
    } else {
       
       if (ind_sp == 1){
          
	  /* read in recombination distance from end of file */
	  if (ind_use_map_marker==0) _fgetln(fp); /* list of recombination distances */
          count=0;
          while (itoken(&map_ord_str_cp,iREQUIRED,&this_loc)) {
	    new_order[count] = this_loc;
	    count++;
	  }
	  if (new_order[0] == 1) {
	    /* disease listed first, skip i=0, first distance */
	    if (ind_use_map_marker==0)
	      if (!rtoken(&ln,rREQUIRED,&dist)) 
	        error("insufficient number of recombination distances");
	    for (i=1; i<count-1; i++) {
	      map_order_chk[i-1] = new_order[i]-2;
	      if (ind_use_map_marker==0){
	        if (!rtoken(&ln,rREQUIRED,&dist)) 
	          error("insufficient number of recombination distances");
		if (dist < 0.0)
		  error("recombination distance < 0");
	        map_distance_f[i-1] = dist;
	        if (dist >= 0.50) cM_dists=TRUE;
	      } 
	    }
	    if (count >= 2) {  /* otherwise no marker specified on line 3 */
	      map_order_chk[count-2] = new_order[count-1]-2;
	      num_in_map_order_chk=count-1;
	    }
	    if (ind_use_map_marker==0 && rtoken(&ln,rREQUIRED,&dist))
	      print("Warning: found more map distances than needed (female map)\n");

	  } else {

	    for (i=0; i<count; i++) {
	      if (new_order[i]==1) error("disease locus cannot be in map");
	      map_order_chk[i]=new_order[i]-2;
	    }
	    if (ind_use_map_marker==0){
	      for (i=0; i<count-1; i++) {
	         /* do a better job of conversion here */
	         if (!rtoken(&ln,rREQUIRED,&dist)) 
	           error("insufficient number of recombination distances");
		 if (dist < 0.0)
		   error("recombination distance < 0");
	         map_distance_f[i] = dist;
	        if (dist >= 0.50) cM_dists=TRUE;
	      }
	      if (ind_use_map_marker==0 && rtoken(&ln,rREQUIRED,&dist))
		print("Warning: found more map distances than needed (female map)\n");
	    }
	    num_in_map_order_chk=count;
	  }
       }
    }

    if ((ind_sp == 1) && (num_in_map_order != num_in_map_order_chk))
      error("num_in_map_order and num_in_map_order_chk don't match (locprep.c)");

    if ((ind_use_map_marker == 0) && (num_in_map_order > 1)) {

      /* look at male distances, convert to cM if necessary */
      if (!cM_dists) for (i=0; i<num_in_map_order-1; i++)
	map_distance_m[i] = rec_to_dist(map_distance_m[i]);

      if (ind_sp == 1) {  /* sex-specific case */
	/* look at female distances, convert to cM if necessary */
	if (!cM_dists) for (i=0; i<num_in_map_order-1; i++)
	  map_distance_f[i] = rec_to_dist(map_distance_f[i]);
      } else {  /* sex-averaged case */
	for (i=0; i<num_in_map_order-1; i++)
	  map_distance_f[i] = map_distance_m[i];
      }

      /* set sex-averaged map distances (in any case) */
      for (i=0; i<num_in_map_order-1; i++){
         map_distance[i] = 0.5*(map_distance_f[i] + map_distance_m[i]);
      }

      if (!cM_dists) {
	print("Map distances have been interpreted as recombination fractions.\n");
	print("In the following, genetic distances will be given in cM units.\n");
      } else print("Map distances have been interpreted as genetic distances in cM units.\n");
    }
    
    if (ind_use_map_marker==1){
      if (num_in_map_order == 0) error("no marker order given in line 3 of the linkage marker file");
      /*read marker from specified map and compare with user defined markers*/
      sf(ps,"\nreading %s marker positions from %s ...:\n",
	 map_name, map_file); pr();
      if (strcmp(map_name,"decode")==0){
        get_map_marker_decode(map_name, map_file , &nr_marker, &nr_omitted);
      }else if(strcmp(map_name,"marshfield")==0){
        get_map_marker_marshfield(map_name, map_file, &nr_marker, &nr_omitted);
      }else if(strcmp(map_name,"duffy")==0){
        get_map_marker_duffy(map_name, map_file, &nr_marker, &nr_omitted);
      }else if(strcmp(map_name,"nievergelt")==0){
        get_map_marker_nievergelt(map_name, map_file, &nr_marker, &nr_omitted);
      }else if(strcmp(map_name,"rutgers")==0 || strcmp(map_name,"usermap")==0){
        get_map_marker_rutgers(map_name, map_file, &nr_marker, &nr_omitted);
      }
      
      check_map_marker(first_marker);
      repair_map_marker_positions(first_marker);
      check_user_marker_build_distances(); 
    }else if (num_in_map_order > 0) {
      /* use the concept of supporting marker for position calculation */
      /* anyway and fill supp_marker structs for use in newcombo.c */
      /* (functions setup_lodscores_map_sup and build_sxp_pos) */
      /* also correct distances if necessary */
      fill_marker_structs(map_name, &nr_marker);
      /*just for building list of user markers*/
      check_user_marker_build_distances();
    }  
    unarray(new_order, int);
    unarray(map_order_chk, int);
    unarray(strptr, char);
  
  } m_except {
      when ENDOFILE: 
         sf(ps,"Unexpected end-of-file, only %d of %d markers read\n",
         loci_read,num_loci); pr();
         free_map_structs();
         break;
      default:
         relay_messages;
  }
 
}



void read_linkage_marker(FILE *fp, int loci_read)
{
  int i, n, type, num1, num2;
  char *word1, *word2;
  double rnum, total_freq, new_total_freq;

  run {
    /* Read in the line with marker type and num_alleles */
    _fgetln(fp); 
    word1=get_temp_string();
    word2=get_temp_string(); 
    if ((n=sscanf(ln,"%d %d %s %s",&num1,&num2, word1, word2)) < 2) {
      sf(ps,"bad allele number line '%s'",ln); error(ps);
    }

    type = num1;
    if (type == 3) {            /* Marker info */
 
      if (phenotypes_read || covs_read) {  /* added (K. Strauch 04/2008) */
	sf(ps,"markers must be defined before quantitative phenotypes\n       and covariates (line '%s')",ln); error(ps);
      }

      /* Save marker name */
      if (word1[0] == '#' && n == 4) {
        strcpy(locus[loci_read].name,word2);
      } else {
        sf(word1,"loc%d",loci_read+1); 
        strcpy(locus[loci_read].name,word1);
      }      
     /* total number of alleles per marker for PMOD: */
      locus[loci_read].allele_total = num2;     
      sf(ps,"marker %s has %d allele(s)\n", locus[loci_read].name, locus[loci_read].allele_total); pr();
    
            /* Read in the line with all the allele frequencies */
      _fgetln(fp); 
      total_freq = 0.00;
      for (i=0; i<num2; i++) {
	if (!rtoken(&ln,rREQUIRED,&rnum)) {
	  sf(ps,"not enough allele frequencies for marker %d",loci_read+1);
	  error(ps);
	}
	locus[loci_read].allele_freq[i+1]=rnum;/* alleles are listed 1 to n */
	
/*     sf(ps,"allele frequency of allele %4d (marker %2d) : %f \n", 
       (i+1),(loci_read+1),locus[loci_read].allele_freq[i+1]); 
	pr(); */
		
	total_freq+=rnum;
      }

      /* Save other permanent locus info */
      locus[loci_read].allele_sum = total_freq;

      /* new for GHM v3.0 */
      if ((total_freq < .90001 || total_freq > 1.09999)
	  && norm_allfreq == 0) {
        sf(ps,"allele frequencies for marker %s sum to %.4lf instead of 1.0\n"
	      "       - please check datafile and consider turning 'normalize allfreq on'.",locus[loci_read].name,total_freq);
	error(ps);
      }
      if (total_freq < .99999 || total_freq > 1.00001) {
	sf(ps,"WARNING: allele frequencies for marker %s sum to %.4lf instead of 1.0\n",
	   locus[loci_read].name,total_freq); pr();
        if (norm_allfreq == 1) {
	  /* new option/bugfix for GHM v3.0:
	     if desired normalize allele frequencies when sum is not 1.0 */
	  new_total_freq = 0.0;
          for(i=0; i<locus[loci_read].allele_total;i++) {
	    locus[loci_read].allele_freq[i+1] = locus[loci_read].allele_freq[i+1] / total_freq;
	    new_total_freq += locus[loci_read].allele_freq[i+1];
	    if (debug_out_pv) {
              sf(ps,"new allele frequency for allele %d of marker %s is %.6lf\n",
		 i+1,locus[loci_read].name, locus[loci_read].allele_freq[i+1]); pr();
	    }
          }
      locus[loci_read].allele_sum = new_total_freq;
	  sf(ps,"normalization -> allele frequencies for marker %s now sum to %.4lf\n",locus[loci_read].name, new_total_freq); pr();
        } else {
          print("         - use the option 'normalize allfreq on' to fix this issue.\n");
        }
      }

      num_markers++;
    }

    else if (type == 0) {     /* Phenotype info */

      if (covs_read) {  /* added (K. Strauch 04/2008) */
	sf(ps,"quantitative phenotypes must be defined before covariates\n       and after markers (line '%s')",ln); error(ps);
      }

      /* Save trait name */
      if (word1[0] == '#' && n == 4) {
        strcpy(phenotype[loci_read-num_markers].name,word2);
      } else {
        sf(word1,"phen%d",loci_read-num_markers+1); 
        strcpy(phenotype[loci_read-num_markers].name,word1);
      }      

      _fgetln(fp);  /* Skip the next five lines, which contain data for LINKAGE not needed here. */
      _fgetln(fp);  /* It would be nice to accomodate files which lack these lines. */
      _fgetln(fp);
      _fgetln(fp);
      _fgetln(fp);

      phenotypes_read = TRUE;  /* added (K. Strauch 04/2008) */
      num_phenotypes++;
      if (num_phenotypes > MAX_PHENOTYPES)
	error("number of phenotypes exceeds maximum. Try increasing MAX_PHENOTYPES in npl.h");
    }

    else if (type == 4) {    /* Covariate info */
      covs_read = TRUE;  /* added (K. Strauch 04/2008) */
      num_covs++; 
      if (num_covs > MAX_COVARIATES)
	error("number of covariates exceeds maximum. Try increasing MAX_COVARIATES in npl.h");
    }

    else
      error("all loci must have allele number = 0 (phenotypes), 3 (markers), or 4 (covariates)");


  } on_error {
    relay_messages;
  }
}


void free_map_structs(void)
{
  MARKER *p, *q;

  /* Free the structures allocated by load_markers, only
     called if load_markers cannot finish loading the file given */
  unarray(locus, LOCUS_INFO);
  unarray(phenotype, PHENOTYPE_INFO);
  num_markers = 0;
  num_phenotypes = 0; /* added (K. Strauch 04/2008) */
  num_covs = 0;       /* added (K. Strauch 04/2008) */
  hdestroy();   /* delete hash table */
  if (first_marker) for (p = first_marker; p != NULL; p = q) {
    q = p->next;
    free(p);
  }
  first_marker = NULL;  last_marker = NULL;

}


void init_disease_info(void)
{
	/* create a-priori probabilities for originals' disease alleles */
	/* CALLED ONCE when locus information file loaded */
	/* changed to incorporate imprinting */
	double tot_aff, tot_unaff, pen_sum[4];
	int i;
	/* liability classes are 1 to N, create a dummy liability class 0
	   which will represent unknowns and will have the average of the
	   other penetrances */

	pen_sum[HOM_AFF] = pen_sum[HET_UA] = pen_sum[HET_AU]
		= pen_sum[HOM_UNAFF] = 0.0;
	for (i=1; i<=num_liability_classes; i++) {
		pen_sum[HOM_AFF] += penetrance[i][HOM_AFF];
		pen_sum[HET_UA] += penetrance[i][HET_UA];
		pen_sum[HET_AU] += penetrance[i][HET_AU];
		pen_sum[HOM_UNAFF] += penetrance[i][HOM_UNAFF];
	}
	penetrance[0][HOM_AFF] = pen_sum[HOM_AFF]/(double)num_liability_classes;
	penetrance[0][HET_UA] = pen_sum[HET_UA]/(double)num_liability_classes;
	penetrance[0][HET_AU] = pen_sum[HET_AU]/(double)num_liability_classes;
	penetrance[0][HOM_UNAFF] = pen_sum[HOM_UNAFF]/(double)num_liability_classes;

	for (i=0; i<=num_liability_classes; i++) {
		apriori_disease[i][AFFECTED][HOM_AFF] = 
			affected_freq*affected_freq*penetrance[i][HOM_AFF];
		apriori_disease[i][AFFECTED][HET_UA] = 
			affected_freq*(1.0-affected_freq)*
			(penetrance[i][HET_AU]+penetrance[i][HET_UA])/2.0;
		apriori_disease[i][AFFECTED][HET_AU] =
			affected_freq*(1.0-affected_freq)*
			(penetrance[i][HET_AU]+penetrance[i][HET_UA])/2.0;
		apriori_disease[i][AFFECTED][HOM_UNAFF] = 
			(1.0-affected_freq)*(1.0-affected_freq)*penetrance[i][HOM_UNAFF];

		apriori_disease[i][UNAFFECTED][HOM_AFF] =
			affected_freq*affected_freq*(1.0-penetrance[i][HOM_AFF]);
		apriori_disease[i][UNAFFECTED][HET_UA] =
			affected_freq*(1.0-affected_freq)*
			(1.0-(penetrance[i][HET_AU]+penetrance[i][HET_UA])/2.0);
		apriori_disease[i][UNAFFECTED][HET_AU] =
			affected_freq*(1.0-affected_freq)*
			(1.0-(penetrance[i][HET_AU]+penetrance[i][HET_UA])/2.0);
		apriori_disease[i][UNAFFECTED][HOM_UNAFF] = 
			(1.0-affected_freq)*(1.0-affected_freq)*(1.0-penetrance[i][HOM_UNAFF]);

		apriori_disease[i][UNKNOWN][HOM_AFF] =
			apriori_disease[i][AFFECTED][HOM_AFF]+apriori_disease[i][UNAFFECTED][HOM_AFF];
		apriori_disease[i][UNKNOWN][HET_UA] = apriori_disease[i][AFFECTED][HET_UA]
			+ apriori_disease[i][UNAFFECTED][HET_UA]; 
		apriori_disease[i][UNKNOWN][HET_AU] = apriori_disease[i][AFFECTED][HET_AU]
			+ apriori_disease[i][UNAFFECTED][HET_AU];
		apriori_disease[i][UNKNOWN][HOM_UNAFF] =
			apriori_disease[i][AFFECTED][HOM_UNAFF]+apriori_disease[i][UNAFFECTED][HOM_UNAFF];


		/* apriori_disease for affected and unaffected are not normalized */
		if (!algebra)
		{
			tot_aff = apriori_disease[i][AFFECTED][HOM_AFF]+
				apriori_disease[i][AFFECTED][HET_UA]+
				apriori_disease[i][AFFECTED][HET_AU]+
				apriori_disease[i][AFFECTED][HOM_UNAFF];
			if (tot_aff > 0.00000000000001) {
				apriori_disease[i][AFFECTED][HOM_AFF] /= tot_aff;
				apriori_disease[i][AFFECTED][HET_UA] /= tot_aff;
				apriori_disease[i][AFFECTED][HET_AU] /= tot_aff;
				apriori_disease[i][AFFECTED][HOM_UNAFF] /= tot_aff;
			}

			tot_unaff = apriori_disease[i][UNAFFECTED][HOM_AFF]+
				apriori_disease[i][UNAFFECTED][HET_UA]+
				apriori_disease[i][UNAFFECTED][HET_AU]+
				apriori_disease[i][UNAFFECTED][HOM_UNAFF];
			if (tot_unaff > 0.00000000000001) {
				apriori_disease[i][UNAFFECTED][HOM_AFF] /= tot_unaff;
				apriori_disease[i][UNAFFECTED][HET_UA] /= tot_unaff;
				apriori_disease[i][UNAFFECTED][HET_AU] /= tot_unaff;
				apriori_disease[i][UNAFFECTED][HOM_UNAFF] /= tot_unaff;
			}  /* normalization only needed with numeric calculation mode, MB - 09/13 */
		} 
	} 
}


void get_map_marker_marshfield(char* map_name, char* map_file, int* nr_marker, int* nr_omitted){
get_map_marker(map_name, map_file, nr_marker, nr_omitted);
}

void get_map_marker_decode(char* map_name, char* map_file, int* nr_marker, int* nr_omitted){
get_map_marker(map_name, map_file, nr_marker, nr_omitted);
}

void get_map_marker_duffy(char* map_name, char* map_file, int* nr_marker, int* nr_omitted){
get_map_marker(map_name, map_file, nr_marker, nr_omitted);
}	     	    
	                                     
void get_map_marker_rutgers(char* map_name, char* map_file, int* nr_marker, int* nr_omitted){
get_map_marker(map_name, map_file, nr_marker, nr_omitted);
}

void get_map_marker_nievergelt(char* map_name, char* map_file, int* nr_marker, int* nr_omitted){
get_map_marker(map_name, map_file, nr_marker, nr_omitted);
}

void get_map_marker(char* map_name, char* map_file, int* nr_marker, int *nr_omitted){
  FILE * volatile fp;
  FILE * volatile fp_me;
  char first_header[MAX_NAME_SIZE];
  char chrom_name[MAX_NAME_SIZE];
  char marker_name[MAX_NAME_SIZE];
  char primer_name[MAX_NAME_SIZE]; 
  char gen_loc[MAX_NAME_SIZE];
  char gen_loc_m[MAX_NAME_SIZE];
  char gen_loc_f[MAX_NAME_SIZE];
  char map_feature_file_name[MAX_NAME_SIZE];
  char wrn_msg[2000];
  int accepted, header_read, i_ord;
  int marker_name_OK;
  double dbl_gen_loc_m, dbl_gen_loc_a;
  volatile double dbl_gen_loc_f;
  volatile double dbl_gen_loc_o;
  ENTRY itm_htable;

  run{fp=open_file(map_file,READ);}except_when(CANTOPEN){
    sf(ps,"can't open file %s", map_file);
    error(ps);
  } 
  
  run {
 
    if (!hcreate(MAX_HASHTABLE_SIZE)){
      sf(ps,"could not create hashtable for map data");
      error(ps);
    }	      
    *nr_marker=0;
    *nr_omitted=0;
    header_read = 0;
    i_ord = 0;
    supp_marker = NULL;
    first_marker = NULL;
    last_marker = NULL; 
    sf(map_feature_file_name, "%s", map_name);
    strcat(map_feature_file_name, "_omitted_markers");
    fp_me=open_file(map_feature_file_name,WRITE);
  
    while(1) {

      _fgetln(fp);
	
      /* need to call crunch() only once per variable
         crunch(): despace, _filter, lowercase */
       	   
      if (header_read==0){
	     
        /*Read header line, information may be useful*/
        header_read = 1;	     
        if (!maxstoken(&ln,sREQUIRED,first_header,MAX_NAME_SIZE) ||
	    strcmp (crunch(first_header), "na")==0 ||
	    strcmp (first_header, ".")==0 ||
	    !strstr (first_header, "chr")) {
	  sf(ps,"no header line in mapfile (must start with 'chr')");
	  error(ps);
	}
	   
      }else{	   
	      	 
        accepted = 1;
	marker_name_OK = 1;
        wrn_msg[0] = '\0';
        strcpy(marker_name,"");
        strcpy(primer_name,"");
        strcpy(gen_loc,"");
        strcpy(gen_loc_f,"");
        strcpy(gen_loc_m,"");
	         
        /* Read the name of the chromosome*/
        if (!maxstoken(&ln,sREQUIRED, chrom_name,MAX_NAME_SIZE)||
	    strcmp (crunch(chrom_name), "no")==0 ||
	    strcmp (chrom_name, "na")==0 ||
	    strcmp (chrom_name, ".")==0) {
          sf(wrn_msg,"Warning: no chromosome number\n"); 
          accepted = 0;              
        }
	uppercase(chrom_name);
	    
        /* Read the name of the marker*/
        if (!maxstoken(&ln,sREQUIRED, marker_name,MAX_NAME_SIZE)||
	    strcmp (crunch(marker_name), "no")==0 ||
	    strcmp (marker_name, "na")==0 ||
	    strcmp (marker_name, ".")==0 ||
	    strcmp (marker_name, "unknown")==0) {
          marker_name_OK = 0;
        }  
        
        /* Read the name of the primer*/
        if (!maxstoken(&ln,sREQUIRED, primer_name,MAX_NAME_SIZE)||
	    strcmp (crunch(primer_name), "no")==0 ||
	    strcmp (primer_name, "na")==0 ||
	    strcmp (primer_name, ".")==0 ||
	    strcmp (primer_name, "unknown")==0) {
	  if (marker_name_OK == 0) {
	    strcat(wrn_msg,"Warning: no marker name\n");
	    accepted = 0;
	  }
        }else{
          if (marker_name_OK == 0){
            strcpy(marker_name,primer_name);	
            marker_name_OK = 1;
          }      
        }  
	uppercase(marker_name);
        
        /* Read the genetic location*/
        if (!maxstoken(&ln,sREQUIRED, gen_loc,MAX_NAME_SIZE)||
	    strcmp (crunch(gen_loc), "no")==0 ||
	    strcmp (gen_loc, "na")==0 ||
	    strcmp (gen_loc, ".")==0) {
          strcat(wrn_msg,"Warning: no genetic location\n");
          accepted = 0;
        }else{
          dbl_gen_loc_o = atof(gen_loc);	      
        }	      
        
        /* Read the genetic location for females*/
        if (!maxstoken(&ln,sREQUIRED, gen_loc_f,MAX_NAME_SIZE)||
	    strcmp (crunch(gen_loc_f), "no")==0 ||
	    strcmp (gen_loc_f, "na")==0 ||
	    strcmp (gen_loc_f, ".")==0) {
          strcat(wrn_msg,"Warning: no genetic location for females\n");
          accepted = 0;
        }else{
          dbl_gen_loc_f = atof(gen_loc_f);	      
        }	         
                                 
        /* Read the genetic location for males*/
        if (!maxstoken(&ln,sREQUIRED, gen_loc_m,MAX_NAME_SIZE)||
	    strcmp (crunch(gen_loc_m), "no")==0 ||
	    strcmp (gen_loc_m, "na")==0 ||
	    strcmp (gen_loc_m, ".")==0) {
          strcat(wrn_msg,"Warning: no genetic location for males\n");
          accepted = 0;
        }else{
          dbl_gen_loc_m = atof(gen_loc_m);
        }	
	  	  	    
        if (accepted==1){

          supp_marker = malloc(sizeof(struct marker)); 
          if (!supp_marker){
            sf(ps,"not enough memory to read map data");
            error(ps);
          }
          i_ord++;

          dbl_gen_loc_a = 0.5 * (dbl_gen_loc_m + dbl_gen_loc_f);
          strcpy(supp_marker->map_name, map_name);
          strcpy(supp_marker->chromosome, chrom_name);
          strcpy(supp_marker->marker_name, marker_name);
          supp_marker->gen_pos_o = dbl_gen_loc_o;
          if (ind_sp == 1){
            supp_marker->gen_pos_a = dbl_gen_loc_a;
            supp_marker->gen_pos_m = dbl_gen_loc_m;
            supp_marker->gen_pos_f = dbl_gen_loc_f;
          }else{
            supp_marker->gen_pos_a = dbl_gen_loc_o;
            supp_marker->gen_pos_m = dbl_gen_loc_o;
            supp_marker->gen_pos_f = dbl_gen_loc_o;
          }             
          supp_marker->order_number = i_ord;
          supp_marker->ind_user_marker = 0; 
          supp_marker->next_user_marker = NULL;
          supp_marker->prior_user_marker = NULL; 
          itm_htable.key = supp_marker->marker_name;
          itm_htable.data = (void*)supp_marker;

          if (!hsearch(itm_htable, FIND)){
            if (!hsearch(itm_htable, ENTER)){
	      free(supp_marker);
              sf(ps,"error in building htable for marker map");
              error(ps);
            }else{
              dls_store(supp_marker, &first_marker, &last_marker);	     
              (*nr_marker)++;
            }  
          }else{
            strcat(supp_marker->marker_name,"_2");
            itm_htable.key = supp_marker->marker_name;
	    /** Differing from what is stated in hsearch.h, the call
		hsearch(,ENTER) does NOT replace existing data but
		rather returns a pointer to the already existing
		element without changing.  So we always call 
	        hsearch(,FIND) first to be sure. **/
	    if (!hsearch(itm_htable, FIND)){
	      sf(ps,"marker %s already read, will be saved as %s_2\n\n",
		 marker_name, marker_name); fpr(fp_me);
	      if (!hsearch(itm_htable, ENTER)){
		free(supp_marker);
		sf(ps,"error in building htable for marker map");
		error(ps);
	      }else{
		dls_store(supp_marker, &first_marker, &last_marker);	     
		(*nr_marker)++;
	      }
	    }else{
	      sf(ps,"Warning: marker %s more than twice in map file\n",
		 marker_name); fpr(fp_me);
	      sf(ps,"3rd (or additional) occurrence of marker %s omitted\n\n",
		 marker_name); fpr(fp_me);
	      free(supp_marker);
	      (*nr_omitted)++;
	    }
          }      
        }else{          
          sf(ps, "%s", wrn_msg);fpr(fp_me);
          sf(ps,"Warning: bad input line, marker %s omitted\n\n",
	     marker_name);
          fpr(fp_me);
          (*nr_omitted)++;
        }  	     
      }	
      
    } 
    sf(ps,"number of accepted markers: %d \nThese markers and their positions can be found in the file 'used_map'.\n",*nr_marker); pr(); 
    sf(ps,"number of omitted markers: %d \nThese markers are listed in the file '%s'.\n", *nr_omitted, map_feature_file_name); pr();
    close_file(fp);    fp=NULL;
    close_file(fp_me); fp_me=NULL;
  
  } m_except {
      when SOFTABORT:
	 if (fp != NULL) { close_file(fp); fp=NULL; }
         if (fp_me != NULL) { close_file(fp_me); fp_me=NULL; }
	 free_map_structs();
	 relay;
	 break;
      when ENDOFILE:
	 if (fp != NULL) { close_file(fp); fp=NULL; }
         if (fp_me != NULL) { close_file(fp_me); fp_me=NULL; }
	 sf(ps,"number of accepted markers: %d \nThese markers and their positions can be found in the file 'used_map'.\n",*nr_marker); pr();
	 sf(ps,"number of omitted markers: %d \nThese markers are listed in the file '%s'.\n", *nr_omitted, map_feature_file_name); pr();
	 break;
      default:
	 if (fp != NULL) { close_file(fp); fp=NULL; }
	 if (fp_me != NULL) { close_file(fp_me); fp_me=NULL; }
	 free_map_structs();
	 relay_messages;
  }
}

void fill_marker_structs(char* map_name, int* nr_marker){
  char chrom_name[MAX_NAME_SIZE];
  int i, i_ord, changed;
  double dbl_gen_loc_m, dbl_gen_loc_f, dbl_gen_loc_a, total_m, total_f;
  ENTRY itm_htable; 
  
  if (!hcreate(MAX_HASHTABLE_SIZE)){
    sf(ps,"could not create hashtable for map data");
    error(ps);
  }	      
  
  *nr_marker=0;
  i_ord = 0;
  supp_marker = NULL;
  first_marker = NULL;
  last_marker = NULL;
  
  strcpy(map_name,"used_map");
  strcpy(chrom_name,"0");
    
  total_m = 0.0;
  total_f = 0.0;

  for (i=0; i<num_in_map_order; i++) {
    
      changed = 0;
      dbl_gen_loc_m = total_m;
      dbl_gen_loc_f = total_f;

      if (i < num_in_map_order-1) {   /** leave out last marker **/
	if (map_distance_m[i] <= 0) {
	  map_distance_m[i] = MAP_ADJUST_DIST;
	  changed |= 1;
	}
	if (map_distance_f[i] <= 0) {
	  map_distance_f[i] = MAP_ADJUST_DIST;
	  changed |= 2;
	}
	if (changed) {
	  map_distance[i] = 0.5*(map_distance_m[i] + map_distance_f[i]);
	  if (ind_sp==1) {
	    if (changed & 1) {
	      sf(ps,"Warning: male distance zero between markers %s and %s\n         setting male genetic distance to %.4lf cM\n",
		 locus[map_order[i]].name, locus[map_order[i+1]].name,
		 map_distance_m[i]);
	      pr();
	    }
	    if (changed & 2) {
	      sf(ps,"Warning: female distance zero between markers %s and %s\n         setting female genetic distance to %.4lf cM\n",
		 locus[map_order[i]].name, locus[map_order[i+1]].name,
		 map_distance_f[i]);
	      pr();
	    }
	  } else {  /* ind_sp==0 */
	    sf(ps,"Warning: genetic distance zero between markers %s and %s\n         setting genetic distance to %.4lf cM\n",
	       locus[map_order[i]].name, locus[map_order[i+1]].name,
	       map_distance[i]);
	    pr();
	  }
	}
	total_m += map_distance_m[i];
	total_f += map_distance_f[i];
      }
      
      supp_marker = malloc(sizeof(struct marker)); 
      if (!supp_marker){
        sf(ps,"not enough memory to read map data");
        error(ps);
      }
      i_ord++;
      
      dbl_gen_loc_a = 0.5 * (dbl_gen_loc_m + dbl_gen_loc_f);
      strcpy(supp_marker->map_name, map_name);
      strcpy(supp_marker->chromosome, chrom_name);
      strcpy(supp_marker->marker_name, locus[map_order[i]].name);
      uppercase(crunch(supp_marker->marker_name));
      supp_marker->gen_pos_a = dbl_gen_loc_a;
      supp_marker->gen_pos_m = dbl_gen_loc_m;
      supp_marker->gen_pos_f = dbl_gen_loc_f;
      supp_marker->order_number = i_ord;
      supp_marker->ind_user_marker = 0; 
      supp_marker->next_user_marker = NULL;
      supp_marker->prior_user_marker = NULL; 
      itm_htable.key = supp_marker->marker_name;
      itm_htable.data = (void*)supp_marker;
      if (!hsearch(itm_htable, ENTER)){
	free(supp_marker);
        sf(ps,"error in building htable for marker map");
        error(ps);
      } 
      dls_store(supp_marker, &first_marker, &last_marker);

      (*nr_marker)++;
  }
    
  /* sf(ps,"number of accepted markers: %d\nThese markers and their positions can be found in the file 'user_markers_in_map'\n",*nr_marker); pr(); */
}


void check_user_marker_build_distances(void){
    int i, first;
    char marker_name[MAX_NAME_SIZE];
    MARKER *prev_marker;
    MARKER *act_marker;
    ENTRY  itmt_htable;
    ENTRY *found_htable;
    FILE* fp;
    
    void report_missing_user_marker(void);
    
    fp = open_file("user_markers_in_map",WRITE);
    if (ind_sp==1){
      sf(ps,"chr. marker             avg.pos    f.pos      m.pos\n\n");fpr(fp);
    }else if(ind_sp==0){
      sf(ps,"chr. marker             avg.pos\n\n");fpr(fp);  
    }  
    nr_miss_marker = 0;
    first = 1;
    prev_marker = NULL;
           
    for (i=0; i<num_in_map_order; i++) {
             
       found_htable = NULL;
       act_marker = NULL; 
       strcpy(marker_name, locus[map_order[i]].name);
       itmt_htable.key = uppercase(crunch(marker_name));
       found_htable = hsearch(itmt_htable, FIND);
       if (found_htable) {
         act_marker = (struct marker*)(found_htable->data);
         
         /* check for more than one user marker with the same name */
         if (act_marker->ind_user_marker == 1){
	   sf(ps,"user marker %s more than once in linkage marker file!",
	      act_marker->marker_name);
           error(ps);
         }
         
         /* check if all user markers are on the same chromosome */
         if(i>0 && prev_marker &&
	    strcmp(act_marker->chromosome,prev_marker->chromosome)!=0){
           sf(ps,"last marker checked: %s is on different chromosome", act_marker->marker_name);
           error(ps);
         }
         
         /* mark marker as user marker */
         if(act_marker->ind_user_marker == 0){
           act_marker->ind_user_marker = 1;
         }
  
         if (ind_sp==1){
            sf(ps,"%-4s %-15s  %9.4lf  %9.4lf  %9.4lf\n", act_marker->chromosome,
            act_marker->marker_name, act_marker->gen_pos_a,
            act_marker->gen_pos_f, act_marker->gen_pos_m);fpr(fp);
         }else if(ind_sp==0){
            sf(ps,"%-4s %-15s  %9.4lf\n", act_marker->chromosome,
            act_marker->marker_name, act_marker->gen_pos_a);fpr(fp);
         }
         
         
	 if (i>0 && prev_marker){
	   map_distance_m[i-1] = act_marker->gen_pos_m - prev_marker->gen_pos_m;
	   map_distance_f[i-1] = act_marker->gen_pos_f - prev_marker->gen_pos_f;
	   map_distance[i-1]   = act_marker->gen_pos_a - prev_marker->gen_pos_a;
	 }  
       } 
                    
       if (!act_marker){         
         miss_marker[nr_miss_marker] = map_order[i];
	 nr_miss_marker++;
       }else{
          if (first){
	    prev_marker = act_marker;
	    first = 0;
	  }else{
	    if (act_marker->gen_pos_a >= prev_marker->gen_pos_a ||
	        real_eq(act_marker->gen_pos_a,prev_marker->gen_pos_a)){
	      act_marker->prior_user_marker = prev_marker;
	      (act_marker->prior_user_marker)->next_user_marker = act_marker;
	      prev_marker = act_marker;
	               
            }else{  
	      sf(ps,"marker %s not in correct order in linkage marker file",
		 act_marker->marker_name); error(ps);
            }
          }
      }
      
   }
   close_file(fp); fp=NULL;
   if (nr_miss_marker > 0) report_missing_user_marker();   
}  

 void report_missing_user_marker(void){
   int i;
   
   sf(ps,"\nThe following user markers have not been found in the %s map:\n",
      map_name);
   pr();
   
   for (i=0; i<nr_miss_marker; i++){
      sf(ps,"%s\n",locus[miss_marker[i]].name);
      pr();
   }    
   sf(ps,"Please insert these markers in the map file before continuing.");
   error(ps);
 }
 
 
void check_map_marker(struct marker* first_marker){
   MARKER *act_marker;
   MARKER *prev_marker;
   char *map_feature_file_name;
   FILE *fp;
   
   map_feature_file_name = get_temp_string(); 
   sf(map_feature_file_name, "%s", map_name);
   strcat(map_feature_file_name, "_inconsistent_positions");
   fp=open_file(map_feature_file_name,WRITE);
   
   sf(ps,"In the following, markers will be listed whose original positions are\nnot consistent with their sequence in the map file, i.e., with a male\nor female coordinate (sex-specific case) or averaged coordinate\n(sex-averaged case) smaller than the coordinate of the preceding\nmarker.  The inconsistent coordinate will be corrected by setting it\nto the corresponding coordinate of the preceding marker plus a\nvalue of eps=%.4lfcM.  As a consequence, in the sex-specific case,\nthis also affects the sex-averaged position, which is calculated as\nthe arithmetic mean of the female and male genetic coordinates.\nWhen using a sex-specific map, for each inconsistent marker and its\npreceding marker, four coordinates are reported below; the original\naveraged position taken from the map file ('orig'), the arithmetic\nmean of the female and male positions ('avg'), as well as the\noriginal female and male positions taken from the map file ('female'\nand 'male', respectively).  With a sex-averaged map, only the original\naveraged position taken from the map file ('avg') is mentioned for\neach inconsistent marker and its preceding marker.  Please note that,\nif a marker has a male or female position (sex-specific map) or an\naveraged position (sex-averaged map) identical to the one of the\nprevious marker, its position will be corrected as well, for technical\nreasons.  However, such markers are not listed here, since they do not\nrepresent inconsistencies in the narrow sense.  The corrected positions\nof all aforementioned markers can be found in the 'used_map' file.\n\n", MAP_ADJUST_DIST); fpr(fp);
   
   prev_marker = first_marker; 
   act_marker = prev_marker->next;          
   while (act_marker){ 
   
     if (strcmp(act_marker->chromosome,prev_marker->chromosome)==0){
       if ((act_marker->gen_pos_m < prev_marker->gen_pos_m) || 
           (act_marker->gen_pos_f < prev_marker->gen_pos_f)){
           
          if (ind_sp==1){            
            sf(ps,"inconsistency in the following consecutive markers:\n");fpr(fp);
            sf(ps,"orig., avg., female and male position of %s: %7.4lf %7.4lf %7.4lf %7.4lf\n", prev_marker->marker_name, prev_marker->gen_pos_o, prev_marker->gen_pos_a, prev_marker->gen_pos_f, prev_marker->gen_pos_m);fpr(fp);  	  
            sf(ps,"orig., avg., female and male position of %s: %7.4lf %7.4lf %7.4lf %7.4lf\n\n", act_marker->marker_name, act_marker->gen_pos_o, act_marker->gen_pos_a, act_marker->gen_pos_f, act_marker->gen_pos_m);fpr(fp);  	  
          }else if(ind_sp==0){
            sf(ps,"inconsistency in the following consecutive markers:\n");fpr(fp);
            sf(ps,"avg. position of %s: %7.4lf\n", prev_marker->marker_name, prev_marker->gen_pos_a);fpr(fp);  	  
            sf(ps,"avg. position of %s: %7.4lf\n\n", act_marker->marker_name, act_marker->gen_pos_a);fpr(fp);  
          }   
       }    
     }    
     prev_marker = act_marker;
     act_marker = act_marker->next;          
  }
  close_file(fp); fp=NULL;
}


void repair_map_marker_positions(struct marker* first_marker){
   FILE* fp;
   MARKER *act_marker;
   MARKER *prev_marker;

   prev_marker = first_marker; 
   act_marker = prev_marker->next;       
   while (act_marker){
     
     if (strcmp(act_marker->chromosome,prev_marker->chromosome)==0){
   
        if (act_marker->gen_pos_m <= prev_marker->gen_pos_m){
            act_marker->gen_pos_m = prev_marker->gen_pos_m + MAP_ADJUST_DIST;
        }
        if (act_marker->gen_pos_f <= prev_marker->gen_pos_f){
            act_marker->gen_pos_f = prev_marker->gen_pos_f + MAP_ADJUST_DIST;
        }	
        act_marker->gen_pos_a = 0.5*(act_marker->gen_pos_m + act_marker->gen_pos_f); 
        /*act_marker->order_number = atoi(act_marker->chromosome)*1000.0 + 
	                                act_marker->gen_pos_a;	            
       */
    }
    	     
    prev_marker = act_marker;
    act_marker = act_marker->next;   
    
  }
  
  
  if (ind_use_map_marker==1){
  
    fp = open_file("used_map",WRITE);
    sf(ps,"These are the positions of the markers found in the map file as\nthey are actually used by the program.  In case that a sex-specific\nmap is used, the averaged genetic coordinate for each marker,\ncalculated as the arithmetic mean of the female and male positions\nrather than the original averaged coordinate, is reported as well.\nWhen employing a sex-averaged map, the position corresponds to the\noriginal averaged coordinate.  Positions of markers which were not\nconsistent with their sequence in the map file, i.e., with a male\nor female coordinate (sex-specific case) or averaged coordinate\n(sex-averaged case) smaller than the coordinate of the preceding\nmarker, have been corrected in the actually used map that is reported\nhere.  The correction was done by setting the inconsistent coordinate\nto the corresponding coordinate of the preceding marker plus a value\nof eps=%.4lfcM.  As a consequence, in the sex-specific case, this\nalso affects the sex-averaged position.  The corrected markers, along\nwith their original positions as specified in the map file, can be\nfound in the file '<map>_inconsistent_positions'.  If a marker has a\nmale or female position (sex-specific map) or an averaged position\n(sex-averaged map) identical to the one of the previous marker, its\nposition has been corrected as well, for technical reasons.  Since\nthese markers do not represent inconsistencies in the narrow sense,\nthey are not listed in the file '<map>_inconsistent_positions'.\nPlease see below for the new coordinates of all these markers.\n\n\n", MAP_ADJUST_DIST); fpr(fp);
    
    if (ind_sp==1){
      sf(ps,"chr. marker             avg.pos    f.pos      m.pos\n\n");fpr(fp);
      act_marker = first_marker;
      while (act_marker){
        sf(ps,"%-4s %-15s  %9.4lf  %9.4lf  %9.4lf\n", act_marker->chromosome,
            act_marker->marker_name, act_marker->gen_pos_a,
            act_marker->gen_pos_f, act_marker->gen_pos_m);fpr(fp); 
        act_marker = act_marker->next;
      }
    }else if (ind_sp==0){
      sf(ps,"chr. marker             avg.pos\n\n");fpr(fp);
      act_marker = first_marker;
      while (act_marker){
        sf(ps,"%-4s %-15s  %9.4lf\n", act_marker->chromosome,
            act_marker->marker_name, act_marker->gen_pos_a);fpr(fp); 
        act_marker = act_marker->next;
      }  
    }
  
    close_file(fp); fp=NULL;
  
  }
  
  
}


/* Create a doubly linked list in sorted order. */
void dls_store(struct marker *i, struct marker **start, struct marker **last)
{
  struct marker *old, *p;

  if(*last==NULL) { /* first element in list */
    i->next = NULL;
    i->prior = NULL;
    *last = i;
    *start = i;
    return;
   }

  p = *start; /* start at top of list */

  old = NULL;
  while(p) {
    /*if(strcmp(p->name, i->name)<0){*/
     if( p->order_number <= i->order_number){   
      old = p;
      p = p->next;
    }
    else {
      if(p->prior) {
        p->prior->next = i;
        i->next = p;
        i->prior = p->prior;
        p->prior = i;
        return;
      }
      i->next = p; /* new first element */
      i->prior = NULL;
      p->prior = i;
      *start = i;
      return;
    }
  }
  old->next = i; /* put on end */
  i->next = NULL;
  i->prior = old;
  *last = i;
}
