/* GENEHUNTER-MODSCORE - pmod.c
   module to perform p-value calculation for MOD-score and LOD-score analysis
   written by Manuel Mattheisen, Copyright (c) 2005-2019 */

#define INC_LIB
#define INC_SHELL
#include "npl.h"
#include "dSFMT.h"   /* Mersenne Twister */
#include "time.h"
#include <strings.h>

void init_disease_info(void);
void analyze_family(volatile int num_in_ped);

/* for PMOD */
#define real_eq(x,y) (fabs((x)-(y))<1e-10)
#define real_map_eq(x,y) (fabs((x)-(y))<1e-3)
#define rand() (z=z*1103515245+12345, z&0x7fff) 
#define srand(prnum_g_s) (z=prnum_g_s)
#define prnum_g_x(prnum_g_x) (ix = prnum_g_x)
#define REPLITICK 10

#define MISSING (-1) /* Value to set uncalculated fields to */
#define MALE 1
#define FEMALE 2

/* For Debugging and Testing, uncomment as necessary:

#define PMOD1

*/

/* local methods */
void pmod_linkage_pedigree(void);
void simulate_alleles(void);
void report_values(void);
void modi_allele_freq(void);
void retouch_sim_alleles(void);
void sim_markers_status(int, int);  /*for debug only*/

/* in finalprep.c */
void set_mom_index(int,int);
void set_dad_index(int,int);
void algebra_power_prep(void); /* for algebraic calculus */

static int list_index;

static int total_meioses;

SAVE_PED_MEMBER **save_famtree;         /*** PMOD ***/
int save_list_index[MAX_PEDIGREES];     /*** PMOD ***/
int save_num_pedigrees;                 /*** PMOD ***/
bool fill_save_famtree;                 /*** PMOD ***/
int skip_count;                         /*** PMOD ***/

int replicates_count;                   /*** PMOD ***/
static double *save_total_lod;          /*** PMOD ***/
static int *exceed_total_lod;           /*** PMOD ***/
static int **exceed_total_lod_grid;     /*** PMOD ***/
static double max_total_lod;            /*** PMOD ***/
static int max_pos;                     /*** PMOD ***/
int save_max_pos;                       /*** PMOD ***/
static double save_max_total_lod;       /*** PMOD ***/
static int exceed_max_total_lod;        /*** PMOD ***/
static int *exceed_max_total_lod_grid;  /*** PMOD ***/
static int *seq_sim_single_rcount;      /*** PMOD ***/
static int seq_sim_rcount;              /*** PMOD ***/
double gridpoint_pmod;                  /*** PMOD ***/
static int exceed_count;                /*** PMOD ***/
static int seq_sim_reached;             /*** PMOD ***/
static int seq_sim_max;                 /*** PMOD ***/
static int seq_sim_maxpos;              /*** PMOD ***/

unsigned long int z=1;                  /*** PMOD ***/
int ix;                                 /*** PMOD ***/
double pseu_rn;                         /*** PMOD ***/

bool allele_sum_flag;                   /*** PMOD ***/
bool pmod_calc_flag;                    /*** PMOD ***/

char rfile_count[10];                   /*** PMOD ***/

bool display_scores_old;                /*** PMOD ***/
double screen_count;                    /*** PMOD ***/
static int anatype_old;                 /*** PMOD ***/
static int hapout_old;                  /*** PMOD ***/
static int ser_old;                     /*** PMOD ***/
static int postout_old;                 /*** PMOD ***/
static int dumpre_old;                  /*** PMOD ***/
static int compsha_old;                 /*** PMOD ***/
static int longmod_old;                 /*** PMOD ***/

FILE *pmodfp;                           /*** PMOD ***/
FILE *replifp;                          /*** PMOD ***/
FILE *replifp2;                          /*** PMOD ***/

bool signi_reached;                     /*** PMOD ***/
bool simulation_alleles_done;           /*** PMOD ***/
int **pseudo_haplotype_allele;          /*** PMOD ***/


char allocate_flip_mask[MAX_PEDIGREES]; /*** dealloc. for PMOD, used in finalprep.c ***/
char allocate_maskTest[MAX_PEDIGREES];  /*** dealloc. for PMOD, used in finalprep.c ***/
char allocate_pvct[MAX_PEDIGREES];      /*** dealloc. for PMOD, used in finalprep.c ***/


static char current_ped[MAX_NAME_SIZE];
char save_current_ped[MAX_NAME_SIZE];   /*** PMOD ***/
static int num_allele_data[MAX_INDIVIDUALS];

command pmod_calc(void)
{
  char *filename, *repli_filename, *repli_filename2;
  void pmod_linkage_pedigree(void); 
  int i,j;

  if (replicates_count > 0) {
    print("Since p-values have already been calculated, this command will be ignored.\n");
    return;
  }

  if (mobit_simulate) {
    print("Cannot execute 'calculate p value' command after 'mobit simulation' command.  These commands are mutually exclusive.\n");
    return;
  }

  srand(time(NULL));
  prnum_g_x(rand());

  allele_sum_flag = FALSE;
  simulation_alleles_done = FALSE;

  replicates_count = 0;
  max_total_lod = -999.0;
  save_max_total_lod = -999.0;
  exceed_max_total_lod = 0;
  exceed_count = 0;
  screen_count = 1;
  famnum_count = 0;

  if (analysis_type == NPL_ANALYSIS) {
    sf(ps,"The p-value calculation only works if 'analysis' ist set to\n"
          "'LOD' or 'BOTH'.");
    error(ps);
  } else if (best_position && !modscore_done) {
    sf(ps,"The best position option only works with MOD-score analysis using the\n'modcalc single' option.");
    error(ps);
  } else if (best_position && modcalc == 1) {
    sf(ps,"The best position option only works with 'modcalc single'\nor single-model LOD-score analysis.");
    error(ps);
  } else if (replicates_error_flag || scan_error_flag) {
    sf(ps,"During the analysis of pedigrees, one or more problem(s) arose.\n"
          "Cannot run p-value calculation!");
    error(ps);
  } else if (modcalc && !modscore_done) {
    sf(ps,"You specified the 'modcalc' option, but did not run 'modscore'.\n"
          "Please run 'modscore' first in order to obtain a p-value for\n"
          "the MOD score(s) or discard the 'modcalc' option.");
    error(ps);
  } else if (!scan_done) {
    sf(ps, "No pedigrees were analyzed so far. You must call 'scan pedigrees'\n"
           "before you can calculate p-values for parametric scores.");
    error(ps);
  } else if (num_replicates < 1) {
    sf(ps, "You must specify the desired number of replicates.\n"
           "Use the 'number of replicates' option before calling 'scan pedigrees'.");
    error(ps);
  } else if (changed_modscore_options) {
    sf(ps, "You have executed the 'modscore' command more than once\n"
            "or changed one of the options related to MOD-score analysis.\n"
           "Please perform at most one MOD-score analysis, and do not\n"
           "change the corresponding options after calling 'modscore',\n"
           "in case that p-values should be calculated.");
    error(ps);
  } else if (single_point_mode) {
    sf(ps,"The p-value calculation does not work with 'single point mode'.");
    error(ps);
  } else if (show_distribution && sequential_sim != 0) {
    sf(ps,"Reporting of score distributions for replicates does not work\n"
          "with sequential simulation mode. Please discard one of the two options.");
    error(ps);
  }

  filename = get_temp_string();
  use_uncrunched_args();
  get_one_arg(stoken,"",filename);

  if (nullstr(filename)) {
    sf(ps,"You must specify a filename for storing information about the\n"
          "simulation process and results.");
    error(ps);
  }

  pmodfp = open_file(filename,"w");

  pmod_calc_flag = TRUE;

  if (analysis_type != LOD_ANALYSIS) {
    sf(ps,"WARNING: The p-value calculation will only be done for the parametric\n"
          "linkage scores. For this reason, the 'analysis' option will be set\n"
          "to 'LOD' during the p-value calculation.\n");
    pr(); fpr(pmodfp);
  }

  sf(ps,"\nRandom seed for Mersenne Twister :  %d\n", ix); pr(); fpr(pmodfp);

  for (i=0; i<num_in_map_order; i++) {
    if (locus[map_order[i]].allele_sum < 0.9999 
        || locus[map_order[i]].allele_sum > 1.0001) {
      sf(ps,"\nallele frequencies for marker %d sum to %1.6lf instead "
            "of 1.0",map_order[i]+1,locus[map_order[i]].allele_sum);
      pr(); fpr(pmodfp);
      allele_sum_flag = TRUE;
    }
  }

  if (allele_sum_flag) {
    sf(ps,"WARNING: It is recommended to activate the 'normalize allfreq' option!");
    pr();
  }

  nl(); fnl(pmodfp);

  array(save_total_lod, num_w_pos, double)
	/* contains score of the real data set per position */
  array(exceed_total_lod, num_w_pos, int);
	/* contains number of replicates that exceed scores per position */
  array(seq_sim_single_rcount, num_w_pos, int);
	/* contains number of replicates that were necessary to reach
           seq. sim. criteria (position specific 'n') */
  matrix(exceed_total_lod_grid, num_w_pos, 101, int);
	/* contains number of replicates that fall within a certain score intervall per position
           (seperated in a grid from 0 to 10 with intervals of lenght 0.1) */
  array(exceed_max_total_lod_grid, 101, int);
	/* same for real data max */
  array(seq_sim_single, num_w_pos, int);
	/* is 1 for each position that reached the seq. sim. criteria */


  for (i=0; i<101; i++) {
    exceed_max_total_lod_grid[i] = 0;
    for (j=0; j <num_w_pos; j++) {
      exceed_total_lod_grid[j][i] = 0;
      exceed_total_lod[j] = 0;
      seq_sim_single_rcount[j] = 0;
    }
  }

  /* save replicates_count for real data max score: */
  seq_sim_rcount = 0; 
  seq_sim_max = 0;
  seq_sim_maxpos = 0;

  for (i=first_mod_pos; i <=last_mod_pos; i++) {
    seq_sim_single[i] = 0;
    save_total_lod[i] = total_lod[i]; /* save real data scores */
    if (total_lod[i] > max_total_lod) {
      max_total_lod = total_lod[i];
      max_pos = i;
    } /* detect real data maximum score. cave: in case two or more
         positions share the highest score only the first will be considered */
  }

  fnl(pmodfp);
  save_max_total_lod = max_total_lod; /* save maximum real data score */
  save_max_pos = max_pos;             /* save position with maximum real data score*/
  max_total_lod = -999.0;

  if (best_position) {
    sf(ps,"Since 'best position' is activated, the point-wise p-value\n"
          "calculation will be executed only for position %.2lf cM,\n"
          "i.e., the position with the highest score in the 'real data'\n"
          "analysis. No region-wide p-value will be calculated.\n",
           wpos[save_max_pos]);
    pr(); fpr(pmodfp); 
    nl(); fnl(pmodfp);
    first_mod_pos = last_mod_pos = save_max_pos;
  }

  total_meioses=0;

  if (replicates_store_result && debug_out_pv) {
    display_scores_old = display_scores; display_scores = TRUE;
    longmod_old = long_mod_output; long_mod_output = TRUE;
  } else {
    display_scores_old = display_scores; display_scores = FALSE;
    longmod_old = long_mod_output; long_mod_output = FALSE;
  }

  /* Remember some of the options that will not be used during p-value calculation */
  anatype_old = analysis_type; analysis_type = LOD_ANALYSIS;
  ser_old = show_expected_recs; show_expected_recs=FALSE;
  hapout_old = haplotype_output; haplotype_output=FALSE;
  postout_old = postscript_output; postscript_output = FALSE;
  dumpre_old = dump_requirements; dump_requirements = FALSE;
  compsha_old = compute_sharing; compute_sharing = FALSE;

  sf(ps,"\nSimulation is running now...\n"); pr(); fpr(pmodfp);

  /* begin replicate-loop */ 
  do {
    /* remember old model parameters */
    affected_freq = save_affected_freq;
    for (i=0; i<num_liability_classes; i++) {
      penetrance[i+1][HOM_UNAFF] = save_penetrance[i+1][HOM_UNAFF];
      penetrance[i+1][HET_AU] = save_penetrance[i+1][HET_AU];
      penetrance[i+1][HET_UA] = save_penetrance[i+1][HET_UA];
      penetrance[i+1][HOM_AFF] = save_penetrance[i+1][HOM_AFF];
    }

    /* again store some disease probabilities */
    init_disease_info();
    if (replicates_store_result) {
      repli_filename = get_temp_string();
      strcpy(repli_filename, "repli");
      sf(rfile_count,"%d", replicates_count+1);
      strcat(repli_filename, rfile_count);
      strcat(repli_filename,".out");
      replifp = open_file(repli_filename,"w");
    }

    if (replicates_store_pre) {
      repli_filename2 = get_temp_string();
      strcpy(repli_filename2, "repli");
      sf(rfile_count,"%d", replicates_count+1);
      strcat(repli_filename2, rfile_count);
      strcat(repli_filename2,".pre");
      replifp2 = open_file(repli_filename2,"w");
    }

    pmod_linkage_pedigree(); /* fill array and do datafile-model analysis */

    if (!fill_save_famtree || !simulation_alleles_done) return;

   /* in case simulated data only contains 'uninformative pedigrees' 
      set all scores = -9999.0 and skip modscore() */
    if (pedigrees_calculated == 0) {
      for (j=first_mod_pos; j<=last_mod_pos; j++)
        total_lod[j] =  -9999.0;
    } else {
      if (modcalc) 
        modscore();
    }
    max_total_lod = -999.0;

    for (i=first_mod_pos; i<=last_mod_pos; i++) {
      if ((sequential_sim == 0) || 
          (sequential_sim != 0 && modcalc != 2)) {
        /* In case seq. sim is deactivated or is used toether with
           'modcalc global' or single-model LOD-score analysis,
           respectively, overall maximum score of replicate
           is detected considering all positions */
        if (total_lod[i] > max_total_lod) {
         max_total_lod = total_lod[i];
         max_pos = i;
        }
     } else { 
        /* In case seq. sim is used together with 'modcalc single'
           the overall maximum score of the replicate is detected
           considering only the positions that do not reach the
           seq. sim. criterion up to now. Only needed for some
           basic screen output and not for calculation of 
           region-wide p-value */
        if ((total_lod[i] > max_total_lod) && seq_sim_single[i] != 1) {
         max_total_lod = total_lod[i];
         max_pos = i;
        }
      }
    }

    if ((replicates_count%REPLITICK) == 0) {
      printf("\r%6.2lf percent (%d replicates) of the desired %d replicates done", 
           ((double) replicates_count/ (double) num_replicates * 100.0),
             replicates_count, num_replicates);
    }

    if (skipped_peds == 1) {
      sf(ps,"Skipped pedigree in replicate %d: %s\n",
            replicates_count, skipped_peds_char); 
      fpr(pmodfp);
  	  } else if (skipped_peds > 1) {
      sf(ps,"Skipped pedigrees in replicate %d: %s\n", 
            replicates_count, skipped_peds_char);
      fpr(pmodfp);
    }

    if (pedigrees_calculated > 0) {
      if (modcalc) {
        if (ind_sp == 0)
          sf(ps,"MOD score in replicate %4d: %10.6lf  at position %7.2lf\n", 
                 replicates_count, max_total_lod, wpos[max_pos]);
        else
          sf(ps,"MOD score in replicate %4d: %10.6lf  at position %7.2lf (%7.2lf, %7.2lf)\n", 
                 replicates_count, max_total_lod, 0.5*(wpos_m[max_pos]+wpos_f[max_pos]),
                 wpos_f[max_pos], wpos_m[max_pos]);
        fpr(pmodfp);
      } else {
        if (ind_sp == 0)
          sf(ps,"Maximum LOD score in replicate %4d: %10.6lf  at position %7.2lf\n", 
                 replicates_count, max_total_lod, wpos[max_pos]);
        else
          sf(ps,"Maximum LOD score in replicate %4d: %10.6lf  at position %7.2lf (%7.2lf, %7.2lf)\n", 
                 replicates_count, max_total_lod, 0.5*(wpos_m[max_pos]+wpos_f[max_pos]),
                 wpos_f[max_pos], wpos_m[max_pos]);
        fpr(pmodfp);
      }
    }

    /* grid from 0.0 to 10.0 - overall maximum */
    gridpoint_pmod = 0.0;
    for (i=0; i<100; i++) {
      if ((max_total_lod >= gridpoint_pmod) && 
          (max_total_lod < (gridpoint_pmod + 0.10000000000))) {
        exceed_max_total_lod_grid[i]++;
       break;
      }
      gridpoint_pmod += 0.1; 
    }

    if (max_total_lod >= 10.0)
      exceed_max_total_lod_grid[100]++;

    gridpoint_pmod = 0.0;
    max_total_lod = -999.0;

    /* grid from 0.0 to 10.0 - all positions */
    for (j=first_mod_pos; j<=last_mod_pos; j++) {
      for (i=0; i<100; i++) {
        if ((total_lod[j] >= gridpoint_pmod) && 
            (total_lod[j] < (gridpoint_pmod + 0.10000000000))) {
          exceed_total_lod_grid[j][i]++;
          break;
        }
        gridpoint_pmod += 0.1;
      }

      if (total_lod[j] >= 10.0)
        exceed_total_lod_grid[j][100]++;
      gridpoint_pmod = 0.0;
    }
    /* exceed real data result at positions? */
    for (i=first_mod_pos; i<=last_mod_pos; i++) {
      if (modcalc == 2 && sequential_sim != 0) {
        if (!seq_sim_single[i]) {
          if ((total_lod[i] > save_total_lod[i]) ||
              (real_eq(total_lod[i],save_total_lod[i])))
            exceed_total_lod[i]++;
        }
      } else { 
        if ((total_lod[i] > save_total_lod[i]) || 
            (real_eq(total_lod[i],save_total_lod[i])))
          exceed_total_lod[i]++;
        if (total_lod[i] > max_total_lod)
          max_total_lod = total_lod[i];
      }
    }

    /* exceed maximum real data result? */
    if ((max_total_lod > save_max_total_lod) || 
        (real_eq(max_total_lod,save_max_total_lod)))
      exceed_max_total_lod++;

    max_total_lod = -999.0;
    /* sequential simulation stuff */
    if (sequential_sim != 0) {
      if (modcalc == 2 || !modcalc) {
        /* for all positions: */
        for (i=first_mod_pos; i<=last_mod_pos; i++)
          if ((exceed_total_lod[i] >= sequential_sim) 
               && (!seq_sim_single[i])) {
            /* there is no gain in computation time and therefore all replicates are computed for
               all positions with !modcalc */
            seq_sim_single[i] = 1;
            /* assignment is only temporarily for !modcalc and will finally be done later on */
            seq_sim_single_rcount[i] = replicates_count;
            if (modcalc == 2 && pedigrees_calculated > 0) {
              sf(ps,"Seq-Sim p-value for position %7.2lf cM (MOD: %.6lf) is %d / %d = %lf.\n",
                    wpos[i], save_total_lod[i], exceed_total_lod[i], seq_sim_single_rcount[i],
                    (double) exceed_total_lod[i]/ (double) seq_sim_single_rcount[i]);
              fpr(pmodfp);
            }
          }
        seq_sim_reached = TRUE;
        for (i=first_mod_pos; i<=last_mod_pos; i++) 
           if (!seq_sim_single[i])
             seq_sim_reached = FALSE;
        if (!modcalc) {
          /* for overall best score in real data set (only count, no decision with regard to "seq-sim"): */
          if (exceed_max_total_lod >= sequential_sim)
            seq_sim_rcount = replicates_count;
        }
      } else if (modcalc == 1) {
        /* for overall best score in real data set: */
        if (exceed_max_total_lod >= sequential_sim) {
          seq_sim_rcount = replicates_count;
          seq_sim_reached = TRUE;
        } else {
          seq_sim_reached = FALSE;
        }
      }
    }

    if (pedigrees_calculated == 0 && replicates_store_result) {
      sf(ps,"\nThere was no 'informative' pedigree in the simulated data,\nhence the result for all positions was set to -9999.\n");
      fpr(replifp);
    }

    /* end of sequential simulation stuff */
    if (replicates_store_result)
      close_file(replifp);
    if (replicates_store_pre)
      close_file(replifp2);

    if ((replicates_count < num_replicates) && !seq_sim_reached) {

      unarray(total_recs, double);
      unarray(wpos, double);
      unarray(wpos_m, double);
      unarray(wpos_f, double);
      unarray(map_position, double);
      unarray(map_position_m, double);
      unarray(map_position_f, double);
    }
    if (sequential_sim != 0 && seq_sim_reached) {
      sf(ps,"\n\nSince the criterion for sequential simulation (%d) has been\n"
            "reached after %d replicates, p-value calculation\nwill be stopped here.\n", sequential_sim, replicates_count); pr();
      break;
    }

    pedigrees_calculated = 0;
  } while (replicates_count < num_replicates);
  /* end of replicate-loop */


  if (sequential_sim != 0 && seq_sim_reached == FALSE) {
    for (i=first_mod_pos; i<=last_mod_pos; i++)
      if (seq_sim_single_rcount[i] == 0)
        seq_sim_single_rcount[i] = replicates_count;
    if (seq_sim_rcount == 0)
        seq_sim_rcount = replicates_count;
  }

  report_values();

  pmod_calc_flag = FALSE;
  display_scores = display_scores_old;
  analysis_type = anatype_old;
  show_expected_recs = ser_old;
  haplotype_output = hapout_old;
  postscript_output = postout_old;
  dump_requirements = dumpre_old;
  compute_sharing = compsha_old;
  long_mod_output = longmod_old;

  unarray(save_total_lod, double);
  unarray(exceed_total_lod, int);
  unarray(seq_sim_single_rcount, int);
  unmatrix(exceed_total_lod_grid, num_w_pos, int);
  unarray(exceed_max_total_lod_grid, int);
  unarray(seq_sim_single, int);

  simulation_alleles_done = FALSE;
  unmatrix(save_famtree,num_pedigrees_allocated,SAVE_PED_MEMBER);

  close_file(pmodfp);
}

void pmod_linkage_pedigree(void)
{  
  /* derived from prep_linkage_pedigree */
  int i, j;

  strcpy(save_current_ped,"");
  strcpy(skipped_peds_char,"");
  skipped_peds=0;

  if (!fill_save_famtree) {
    sf(ps,"Could not save pedigree structure in save_famtree!"); 
    replicates_error_flag = TRUE;
    error(ps);
  }

  for (famnum=0; famnum<save_num_pedigrees; famnum++) {
    if (allocate_flip_mask[famnum] == 1) {
      unarray (save_flip_mask[famnum], int);
      allocate_flip_mask[famnum] = 0;
    }
    if (allocate_maskTest[famnum] == 1) {
      unmatrix (save_maskTest[famnum], save_f_num_in_map_order[famnum], int);
      allocate_maskTest[famnum] = 0;
    }
    if (allocate_pvct[famnum] == 1) {
      unarray (save_pvct[famnum], double);
      allocate_pvct[famnum] = 0;
    }
    if (modcalc==1) unmatrix(save_p_all_pos[famnum], num_mod_pos, double);
    if (modcalc==2) unarray(save_p_one_pos[famnum], double); 
  }
  unmatrix(famtree,num_pedigrees_allocated, PED_MEMBER);
  matrix(famtree,num_pedigrees_allocated, MAX_INDIVIDUALS+5, PED_MEMBER);

  num_pedigrees = 1; /* num_pedigrees is 1 for the first call of analyze_family in finalprep.c !*/
  famnum = 0;
  datafile_model_incompatible = FALSE;
  list_index = 0;
  if(algebra)
	  algebra_power_prep(); /* prepare powers for datfile model in case family hasn´t been analyzed yet; calculation takes place in analyze_family */ 
  for (famnum_count=0; famnum_count<num_pedigrees_allocated; famnum_count++){
     famnum_pointer[famnum] = famnum_count; 
     algebra_pointer[famnum] = famnum_count;

  /*debug:sf(ps,"famnum_count:%d, famnum_pointer[%d]:%d, num_pedigrees:%d\n", 
         famnum_count, famnum, famnum_pointer[famnum], num_pedigrees); pr(); */

    for (list_index=0;list_index<save_list_index[famnum_count];list_index++){
      strcpy(save_current_ped,
             save_famtree[famnum_count][list_index].current_ped);
      strcpy(famtree[famnum][list_index].indiv_ID,
             save_famtree[famnum_count][list_index].indiv_ID);
      strcpy(famtree[famnum][list_index].dad,
             save_famtree[famnum_count][list_index].dad);
      strcpy(famtree[famnum][list_index].mom,
             save_famtree[famnum_count][list_index].mom);
      famtree[famnum][list_index].sex 
           = save_famtree[famnum_count][list_index].sex;
      allele_data[list_index][0] 
           = save_famtree[famnum_count][list_index].affection_status;
      if (allele_offset ==2)
        allele_data[list_index][1]
           = save_famtree[famnum_count][list_index].liab_class;
      num_allele_data[list_index] 
           = save_famtree[famnum_count][list_index].num_allele_data;
      famtree[famnum][list_index].nkids=0;
      famtree[famnum][list_index].dad_index=MISSING;
      famtree[famnum][list_index].mom_index=MISSING;
      for (j=0; j<MAX_PHENOTYPES; j++)
        famtree[famnum][list_index].phenotype[j] = MISSING_PHENO;
      for (j=0; j<MAX_COVARIATES; j++)
        famtree[famnum][list_index].covariate[j] = 0.0;
    }

    #ifdef PMOD1
      sf(ps,"1 famnum: %d\n",famnum); pr();
      sf(ps,"1 pedigree: %s\n",save_current_ped); pr();
      sf(ps,"1 num_pedigree: %d\n",num_pedigrees); pr();
      sf(ps,"1 save_num_pedigree: %d\n",save_num_pedigrees); pr();
      sf(ps,"1 famnum_count: %d\n",famnum_count); pr();
      sf(ps,"1 num_pedigrees_allocated: %d\n",num_pedigrees_allocated); pr();
    #endif

    simulate_alleles();
    if (!simulation_alleles_done) {
      replicates_error_flag = TRUE;
      sf(ps,"Could not simulate alleles of pedigree %s", current_ped);
      error(ps);
    }
    analyze_family(list_index);

    if (famnum_count != (num_pedigrees_allocated - 1)) {
      num_pedigrees++;
      famnum = num_pedigrees - 1;
    }
  } /* end: for (famnum_count ... */

  /* If desired give result of datafile-model to replicates-file */
  if (replicates_store_result && pedigrees_calculated != 0) {
    sf(ps,"\nHere are the results for the datafile-model:\n");
    fpr(replifp);
    if (ind_sp == 1) {
      sf(ps,"\npos. (avg. female  male)     LOD\n");
      fpr(replifp);
      for (i=first_mod_pos; i<=last_mod_pos; i++) {
          sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf\n", 
                 0.5*(wpos_m[i] + wpos_f[i]), 
                 wpos_f[i], wpos_m[i], save_total_lod[i]);
      fpr(replifp);
      }
    } else {
      sf(ps,"\nposition  LOD\n");
      fpr(replifp);
      for (i=first_mod_pos; i<=last_mod_pos; i++) {
          sf(ps,"%7.2lf %10.6lf\n", wpos[i], save_total_lod[i]);
          fpr(replifp);
      }
    }
  }

  replicates_count++;
  save_num_pedigrees = num_pedigrees;

  #ifdef PMOD1 /* Print save_famtree on screen: */
    for (famnum_count=0; famnum_count<num_pedigrees_allocated; famnum_count++){
      for (list_index=0;list_index<save_list_index[famnum_count];list_index++){
	 sf(ps,"%s ",save_famtree[famnum_count][list_index].current_ped); 
	 pr();
	 sf(ps,"%s ",save_famtree[famnum_count][list_index].indiv_ID); 
	 pr(),
	 sf(ps,"%s ",save_famtree[famnum_count][list_index].dad); 
	 pr();
         sf(ps,"%s ",save_famtree[famnum_count][list_index].mom); 
	 pr();
         sf(ps,"%d ",save_famtree[famnum_count][list_index].sex); 
	 pr();
         sf(ps,"%d ",save_famtree[famnum_count][list_index].affection_status); 
	 pr();
	 if (allele_offset ==2) {
	   sf(ps,"%s ",save_famtree[famnum_count][list_index].liab_class);
	   pr();
	 }
         sf(ps,"\n"); pr();
      }
    }
  #endif

}

void simulate_alleles(void)
{ 
  int i, p, q, r, s, w;
  int current_dad, current_mom;
  bool all_simulated;
  double upper, lower;
  double current_map_distance;

  strcpy(current_ped, save_current_ped);

  matrix (pseudo_haplotype_allele, list_index, 2, int);

  /* adapted from analyze family, will only be executed here (not in analyze_family again), 
     for illustration go to finalprep.c */ 
  for (i=0; i<list_index; i++) {
     set_dad_index(i,list_index);
     set_mom_index(i,list_index);
     famtree[famnum][i].analyzed_parent=FALSE;
     famtree[famnum][i].dad_bit_uninf=FALSE;
     famtree[famnum][i].mom_bit_uninf=FALSE;
     famtree[famnum][i].kids_found=FALSE;
     famtree[famnum][i].replaced_mom=FALSE;
     famtree[famnum][i].not_peeled=FALSE;
     famtree[famnum][i].simulated=FALSE;

     pseudo_haplotype_allele[i][0] = -1;
     pseudo_haplotype_allele[i][1] = -1;

    /* founder-check*/
    if ((strcasecmp(famtree[famnum][i].mom, "0")!=0) && 
        (strcasecmp(famtree[famnum][i].dad,"0")!=0)) {
      famtree[famnum][i].founder = FALSE;
    } else {
      famtree[famnum][i].founder = TRUE;
    }
  }

  simulation_alleles_done = FALSE;
  replicates_error_flag = TRUE;

  /* first step: each marker and allele for each founder */
  if (full_marker_inf) {
  /* founder alleles will be simulated from 1 to n: */  
    q = 0;
    for (r=0; r < num_in_map_order; r++) {
      for (p=0; p < list_index; p++) {
	if (famtree[famnum][p].founder) { 
	  for (s=0; s < 2; s++) {
	    q += 1;
	    if (q > MAX_ALLELE_SIZE) {
              unmatrix (pseudo_haplotype_allele, save_list_index[famnum_pointer[famnum]], int);
              sf(ps,"2 * founder > MAX_ALLELE_SIZE, maybe edit npl.h");
	      error(ps);
	    } else {
	      allele_data[p][allele_offset+(2*map_order[r])+s]=q;
	    }
	  }
	  famtree[famnum][p].simulated = TRUE;
	}
      }
    }
  } else if (!full_marker_inf) {
    /* founder alleles will be simulated randomly: */
    for (r=0; r < num_in_map_order; r++) {
      for (s=0; s < 2; s++) {
        for (p=0; p < list_index; p++) {
	  if (famtree[famnum][p].founder) {
            /* take care of the fact that allele-sums do not necessarely 
		sum to 1 for each marker: multiplicate with allele-sum 
		of the markers, i.e., pseudo-random-numbers has the range 
		from 0 to allele-sum */
	    pseu_rn = genrand_open_open() * locus[map_order[r]].allele_sum; 
            lower = upper = 0.0;
	    /* alleles are listed 1 to n */
            for (q=0; q < (locus[map_order[r]].allele_total); q++) {
	      lower = upper;
	      upper += locus[map_order[r]].allele_freq[q+1]; /* CHECK if this must be "q+2"? */
	      if ((pseu_rn >= lower) && (pseu_rn <= upper)) {
                allele_data[p][allele_offset+(2*map_order[r])+s]=(q+1);
		break;
	      } 
            }
            if (q == locus[map_order[r]].allele_total) {
              allele_data[p][allele_offset+(2*map_order[r])+s]=q;
            }
           famtree[famnum][p].simulated = TRUE;
          }
        }
      }
    }
  } else {
    unmatrix (pseudo_haplotype_allele, save_list_index[famnum_pointer[famnum]], int);
    sf(ps,"Could not simulate founder alleles"); 
    error(ps);
  }

  /* second step: each marker and allele for each non-founder */
  do {  
    for (p=0; p < list_index; p++) {
       current_dad = famtree[famnum][p].dad_index;
       current_mom = famtree[famnum][p].mom_index;

       if ((famtree[famnum][p].simulated == 0) && 
           (famtree[famnum][p].founder == 0) && 
	   (famtree[famnum][current_dad].simulated == 1) && 
	   (famtree[famnum][current_mom].simulated == 1)) {
	 /* loop regarding markers */
	 for (r=0; r < num_in_map_order; r++) {
	    if (r == 0) {
	   /* first allele of first marker (from dad)*/
	     pseu_rn = genrand_open_open();
	     if ((pseu_rn >= 0.0) && (pseu_rn < 0.5)) {
	       pseudo_haplotype_allele[p][0] = 0;	       
	       allele_data[p][allele_offset+(2*map_order[r])] 
	           = allele_data[current_dad][allele_offset+(2*map_order[r])];
	     } else if ((pseu_rn >= 0.5) && (pseu_rn <= 1.0)){ 
               pseudo_haplotype_allele[p][0] = 1; 	
	       allele_data[p][allele_offset+(2*map_order[r])] 
	           = allele_data[current_dad][allele_offset+(2*map_order[r])+1];
	     }
	  /* second allele of first marker (from mom) */
	     pseu_rn = genrand_open_open();
             if ((pseu_rn >= 0.0) && (pseu_rn < 0.5)) {
	       pseudo_haplotype_allele[p][1] = 0; 
	       allele_data[p][allele_offset+(2*map_order[r])+1] 
	           = allele_data[current_mom][allele_offset+(2*map_order[r])];
	     } else if ((pseu_rn >= 0.5) && (pseu_rn <= 1.0)){ 
	       pseudo_haplotype_allele[p][1] = 1;
	       allele_data[p][allele_offset+(2*map_order[r])+1] 
	           = allele_data[current_mom][allele_offset+(2*map_order[r])+1];
             }

	     /* if (replicates_store_result && debug_out_pv)
	       sim_markers_status(p,r);  only for debbuging */

	   } else {
	    /* first allele of other markers (from dad)*/
             if (ind_sp == 0)
               current_map_distance = dist_to_rec(map_distance[r-1]);
             else 
               current_map_distance = dist_to_rec(map_distance_m[r-1]);
	     pseu_rn = genrand_open_open();
             if ((pseu_rn >= 0.0) && (pseu_rn <= current_map_distance)) {
	       if (pseudo_haplotype_allele[p][0] == 0) {
	         pseudo_haplotype_allele[p][0] = 1;
	       } else if (pseudo_haplotype_allele[p][0] == 1) {
	         pseudo_haplotype_allele[p][0] = 0;
	       }
             }
	     allele_data[p][allele_offset+(2*map_order[r])+0] 
	         = allele_data[current_dad][allele_offset
		 + (2*map_order[r])
		 + pseudo_haplotype_allele[p][0]];
	  /* second allele of other markers (from mom) */
             if (ind_sp == 0)
               current_map_distance = dist_to_rec(map_distance[r-1]);
             else 
               current_map_distance = dist_to_rec(map_distance_f[r-1]);
	     pseu_rn = genrand_open_open();
	     if ((pseu_rn >= 0.0) && (pseu_rn <= current_map_distance)) {
	       if (pseudo_haplotype_allele[p][1] == 0) {
		 pseudo_haplotype_allele[p][1] = 1;
	       } else if (pseudo_haplotype_allele[p][1] == 1) {
		 pseudo_haplotype_allele[p][1] = 0;
	       }
	     }
	     allele_data[p][allele_offset+(2*map_order[r])+1] 
	         = allele_data[current_mom][allele_offset
		 + (2*map_order[r])
		 + pseudo_haplotype_allele[p][1]];

	     /* if (replicates_store_result && debug_out_pv)
               sim_markers_status(p, r); only for debugging */
	   }
	 } /* end "marker-loop" if indivs parents simulated */
         famtree[famnum][p].simulated = TRUE;
       }
    } /* end "for-indiv-loop" */
    /* check after first round: is everyone simulated? */
    all_simulated = TRUE;
    for (w=0; w < list_index; w++) 
      if ((famtree[famnum][w].simulated) == 0)
        all_simulated = FALSE;
  } while (all_simulated == FALSE);

  if (replicates_store_result && debug_out_pv) {
    sf(ps,"\nAll persons have been simulated!\n"); fpr(replifp);
  }

  simulation_alleles_done = TRUE;
  replicates_error_flag = FALSE;

 /* third step: dependent on "simulate untyped", "untyped founder" and 
                retouch simulated alleles */
  retouch_sim_alleles();

  unmatrix (pseudo_haplotype_allele, save_list_index[famnum_pointer[famnum]], int);
}

void retouch_sim_alleles(void)
{
  int p, r, i, j, k, must_retouch;

  must_retouch = 0;

  for (p=0; p < list_index; p++)
    if (save_famtree[famnum_pointer[famnum]][p].is_genotyped == 0)
      must_retouch++; 

  if (replicates_store_result && debug_out_pv )
    if ((!simulate_untyped && must_retouch > 0) || untyped_founders) {
      sf(ps,"\nHere are the simulated alleles before 'retouching' them:\n"); 
      fpr(replifp);
      for(i=0; i<list_index; i++) {
        sf(ps,"%s ", current_ped); fpr(replifp);
        sf(ps,"%s ", famtree[famnum][i].indiv_ID); fpr(replifp);
        sf(ps,"%s ", famtree[famnum][i].dad); fpr(replifp);
        sf(ps,"%s ", famtree[famnum][i].mom); fpr(replifp);
        sf(ps,"%d ", famtree[famnum][i].sex); fpr(replifp);
        sf(ps,"%d ", allele_data[i][0]); fpr(replifp);
        if(allele_offset == 2) {
          sf(ps, "%d ", allele_data[i][1]); fpr(replifp);
        }
        sf(ps," "); fpr(replifp);
        for (k=0; k < num_in_map_order; k++) {
           for (j=0; j<2; j++) {
              sf(ps,"%d ", allele_data[i][allele_offset+(2*map_order[k])+j]); 
              fpr(replifp);
           }
        }
        sf(ps,"\n"); fpr(replifp);
      } 
    }

  if (!simulate_untyped) {
    for (p=0; p < list_index; p++) {
      if (!(save_famtree[famnum_pointer[famnum]][p].is_genotyped)) { 
        for (r=0; r < num_in_map_order; r++) {
           /* debug: sf(ps,"save_famtree[%d][%d].is_genotyped: %d\n", 
             famnum_pointer[famnum], p, 
	     save_famtree[famnum_pointer[famnum]][p].is_genotyped); 
           pr(); */
	   allele_data[p][allele_offset+(2*map_order[r])+0]=0;
	   allele_data[p][allele_offset+(2*map_order[r])+1]=0;
	}
      } else {
	for (r=0; r < num_in_map_order; r++) {
 	  /* sf(ps,"save_famtree[%d][%d].zero_marker[%d]:", 
	           famnum_pointer[famnum], p, map_order[r]); 
          pr();
	  sf(ps,"%d\n", 
	         save_famtree[famnum_pointer[famnum]][p].zero_marker[map_order[r]]);
          pr(); */
	  if (save_famtree[famnum_pointer[famnum]][p].zero_marker[map_order[r]]==1) {
	    allele_data[p][allele_offset+(2*map_order[r])+0]=0;
	    allele_data[p][allele_offset+(2*map_order[r])+1]=0;
	  }
	}
      }
    }
  }

  if (untyped_founders) {
    for (p=0; p < list_index; p++) {
      if (famtree[famnum][p].founder) { 
        for (r=0; r < num_in_map_order; r++) {
	  allele_data[p][allele_offset+(2*map_order[r])+0]=0;
	  allele_data[p][allele_offset+(2*map_order[r])+1]=0;
	}
      }
    }
  }

  /* if desired store pedigree structure in files */
  if (replicates_store_result && debug_out_pv) {
    if ((!simulate_untyped && must_retouch > 0) || untyped_founders)  {
        sf(ps,"\nHere are the simulated alleles after 'retouching' them:\n"); 
        fpr(replifp);
      for (i=0; i<list_index; i++) {
         sf(ps, "%s ", current_ped); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].indiv_ID); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].dad); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].mom); fpr(replifp);
         sf(ps, "%d ", famtree[famnum][i].sex); fpr(replifp);
         sf(ps, "%d ", allele_data[i][0]); fpr(replifp);
         if (allele_offset == 2) {
	   sf(ps, "%d ", allele_data[i][1]); 
	   fpr(replifp);
         }
         sf(ps, " "); fpr(replifp);
         for (k=0; k<num_in_map_order; k++) {
            for (j=0; j<2; j++) {
	       sf(ps, "%d ", allele_data[i][allele_offset+2*map_order[k]+j]); 
               fpr(replifp);
            }
         }
         sf(ps, "\n"); fpr(replifp);
      }
    } else if ((replicates_store_result && simulate_untyped && !untyped_founders) || (replicates_store_result && !simulate_untyped && must_retouch == 0)) {
        sf(ps,"\nHere are the simulated alleles:\n"); 
        fpr(replifp);
      for (i=0; i<list_index; i++) {
         sf(ps, "%s ", current_ped); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].indiv_ID); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].dad); fpr(replifp);
         sf(ps, "%s ", famtree[famnum][i].mom); fpr(replifp);
         sf(ps, "%d ", famtree[famnum][i].sex); fpr(replifp);
         sf(ps, "%d ", allele_data[i][0]); fpr(replifp);
         if (allele_offset == 2) {
	   sf(ps, "%d ", allele_data[i][1]); 
	   fpr(replifp);
         }
         sf(ps, " "); fpr(replifp);
         for (k=0; k<num_in_map_order; k++) {
            for (j=0; j<2; j++) {
	       sf(ps, "%d ", allele_data[i][allele_offset+2*map_order[k]+j]); 
               fpr(replifp);
            }
         }
         sf(ps, "\n"); fpr(replifp);
      }
    }
  }

  if (replicates_store_pre) {
    for (i=0; i<list_index; i++) {
       sf(ps, "%s ", current_ped); fpr(replifp2);
       sf(ps, "%s ", famtree[famnum][i].indiv_ID); fpr(replifp2);
       sf(ps, "%s ", famtree[famnum][i].dad); fpr(replifp2);
       sf(ps, "%s ", famtree[famnum][i].mom); fpr(replifp2);
       sf(ps, "%d ", famtree[famnum][i].sex); fpr(replifp2);
       sf(ps, "%d ", allele_data[i][0]); fpr(replifp2);
       if (allele_offset == 2) {
	 sf(ps, "%d ", allele_data[i][1]); 
	 fpr(replifp2);
       }
       for (k=0; k<num_in_map_order; k++) {
          for (j=0; j<2; j++) {
	     sf(ps, " %d", allele_data[i][allele_offset+2*map_order[k]+j]); 
             fpr(replifp2);
          }
       }
       sf(ps, "\n"); fpr(replifp2);
    }
  }
}

void sim_markers_status(int p, int r)
{
  int current_dad, current_mom;
  
  current_dad = famtree[famnum][p].dad_index;
  current_mom = famtree[famnum][p].mom_index;

   /* this is only for debug!*/
  sf(ps,"marker: %s, allele1: %d, allele2: %d\n", locus[r].name, 
        allele_data[p][allele_offset+(2*map_order[r])],
        allele_data[p][allele_offset+(2*map_order[r])+1]);
  fpr(replifp);
  sf(ps,"   dad - allele1: %d, allele2: %d\n", 
        allele_data[current_dad][allele_offset+(2*map_order[r])],
        allele_data[current_dad][allele_offset+(2*map_order[r])+1]);
  fpr(replifp);
  sf(ps,"   mom - allele2: %d, allele2: %d\n\n", 
        allele_data[current_mom][allele_offset+(2*map_order[r])],
        allele_data[current_mom][allele_offset+(2*map_order[r])+1]);
  fpr(replifp);
}

void report_values(void)
{
  int i,j;
  int exceed_cummulus = 0;
  bool early_exit_rv = FALSE;
  int pos_exceed = 0;
  int pos_count = 0;

  if (exceed_max_total_lod == 0 && !best_position) {
    if ((modcalc == 2 && sequential_sim == 0) || (modcalc == 1) || !(modcalc)) {
      if ((modcalc == 2 && sequential_sim == 0) || (modcalc == 1)) 
        sf(ps,"\n\nNone of the %d replicates reached the overall real data maximum MOD\n"
              "score of %.6lf.\n",
              replicates_count, save_max_total_lod);
      else if (!modcalc)
        sf(ps,"\n\nNone of the %d replicates reached the overall real data maximum LOD\n"
              "score of %.6lf.\n", 
              replicates_count, save_max_total_lod);
      pr(); fpr(pmodfp);
      /* since none of the replicates at least equaled the real data score, simulation proceeded
         up to the specified number of replicates for sequential_sim as well */
      sf(ps,"\nThe p-value is < %lf (increase number of replicates\n"
            "in order to get a more precise result).\n",
            (double) 1 / (double) replicates_count); 
      pr(); fpr(pmodfp);
    }
  } else if (!best_position) {
    if (sequential_sim != 0) {
      if (modcalc != 2) {
        if (exceed_max_total_lod > 1) {
          if (modcalc)
            sf(ps,"\n\n%d replicates exceeded the overall real data maximum MOD score\n"
                  "of %.6lf. ", exceed_max_total_lod,
                   save_max_total_lod);
          else if (!modcalc)
            sf(ps,"\n\n%d replicates exceeded the overall real data maximum LOD score\n"
                  "of %.6lf. ", exceed_max_total_lod,
                   save_max_total_lod);
        } else if (exceed_max_total_lod == 1) {
          if (modcalc)
            sf(ps,"\n\nOne replicate exceeded the overall real data maximum MOD score\n"
                  "of %.6lf. ", save_max_total_lod);
          else if (!modcalc)
            sf(ps,"\n\nOne replicate exceeded the overall real data maximum LOD score\n"
                  "of %.6lf. ", save_max_total_lod);
        }
        pr(); fpr(pmodfp);
        sf(ps,"For this score the p-value is %lf.\n",
              ((double) exceed_max_total_lod / (double) seq_sim_rcount));
        pr(); fpr(pmodfp);
      }
    } else {
      if (exceed_max_total_lod > 1) {
        if (modcalc)
          sf(ps,"\n\n%d replicates exceeded the overall real data maximum MOD score\n"
                "of %.6lf. ", exceed_max_total_lod,
                 save_max_total_lod);
        else if (!modcalc)
          sf(ps,"\n\n%d replicates exceeded the overall real data maximum LOD score\n"
                "of %.6lf. ", exceed_max_total_lod,
                 save_max_total_lod);
      } else if (exceed_max_total_lod == 1) {
        if (modcalc)
          sf(ps,"\n\nOne replicate exceeded the overall real data maximum MOD score\n"
                "of %.6lf. ", save_max_total_lod);
        else if (!modcalc)
          sf(ps,"\n\nOne replicate exceeded the overall real data maximum LOD score\n"
                "of %.6lf. ", save_max_total_lod);
      }
      pr(); fpr(pmodfp);
      sf(ps,"For this score the p-value is %lf.\n",
            ((double) exceed_max_total_lod / (double) replicates_count));
      pr(); fpr(pmodfp);
    }
  }

  if (((sequential_sim == 0 && modcalc == 2) || (modcalc == 1) || (modcalc == 0)) && !best_position) {
    sf(ps,"\nPlease keep in mind that this p-value is 'region-wide' and\n"
          "therefore still needs to be corrected for the 'genome'.\n");
    pr(); fpr(pmodfp);
  }

  /* Give pointwise p-values for positions (only for modcalc == 2 and single-model LOD score)*/

  if (modcalc != 1 ) {
    for (i=first_mod_pos; i<=last_mod_pos; i++) {
      exceed_cummulus = exceed_cummulus + exceed_total_lod[i];
    }
    if (exceed_cummulus != 0) {
      if (best_position)
        sf(ps,"\nSince the 'best position' option was activated, only the p-value for\n"

              "the position with the best score in the real data set is reported:\n");
      else if (!best_position)
        sf(ps,"\nPointwise p-values per position:\n");
      pr(); fpr(pmodfp);

      if (ind_sp == 1) {
        if (modcalc)
          sf(ps,"\npos. (avg. female  male)     MOD     p-value    (exceed / total)\n");
        else
          sf(ps,"\npos. (avg. female  male)     LOD     p-value    (exceed / total)\n");
        pr(); fpr(pmodfp);
        for (i=first_mod_pos; i<=last_mod_pos; i++) {
          pos_exceed = exceed_total_lod[i];
          if (modcalc) 
            pos_count = seq_sim_single_rcount[i];
          else if (!modcalc)
            pos_count = replicates_count;
          if (sequential_sim != 0)
            sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf   %lf    (%d / %d)\n", 
                   0.5*(wpos_m[i] + wpos_f[i]), 
                   wpos_f[i], wpos_m[i], save_total_lod[i],
                   ((double) pos_exceed / (double) pos_count),
                   pos_exceed, pos_count);
          else if (sequential_sim == 0)
            sf(ps,"%7.2lf %7.2lf %7.2lf %10.6lf   %lf    (%d / %d)\n", 
                   0.5*(wpos_m[i] + wpos_f[i]), 
                   wpos_f[i], wpos_m[i], save_total_lod[i],
                   ((double) pos_exceed / (double) replicates_count),
                   pos_exceed, replicates_count);
          pr(); fpr(pmodfp);
        }
      } else {
        if (modcalc) 
          sf(ps,"\nposition  MOD        p-value    (exceed / total)\n");
        else
          sf(ps,"\nposition  LOD        p-value    (exceed / total)\n");
        pr(); fpr(pmodfp);
        for (i=first_mod_pos; i<=last_mod_pos; i++) {
          pos_exceed = exceed_total_lod[i];
          if (modcalc) 
            pos_count = seq_sim_single_rcount[i];
          else if (!modcalc)
            pos_count = replicates_count;
          if (sequential_sim != 0)
            sf(ps,"%7.2lf %10.6lf   %lf    (%d / %d)\n", wpos[i], save_total_lod[i],
                  ((double) pos_exceed / (double) pos_count),
                  pos_exceed, pos_count);
          else if (sequential_sim == 0)
            sf(ps,"%7.2lf %10.6lf   %lf    (%d / %d)\n", wpos[i], save_total_lod[i],
                  ((double) pos_exceed / (double) replicates_count),
                  pos_exceed, replicates_count);
          pr(); fpr(pmodfp);
        }
      }
    } else {
      if (best_position) {  /* only possible for modcalc single */
	sf(ps,"\nNone of the %d replicates exceeded the real data\n"
	      "MOD score at the best position (%.2lf cM).\n",
	   replicates_count, wpos[save_max_pos]);
	pr(); fpr(pmodfp);
	sf(ps,"\nThe p-value for the position %.2lf cM is < %lf (increase number\n"
	   "of replicates in order to get a more precise result).\n",
	   wpos[save_max_pos], (double) 1 / (double) replicates_count);
	pr(); fpr(pmodfp);
      } else {
	if (modcalc == 2)
	  sf(ps,"\nNone of the %d replicates exceeded the real data\n"
	        "MOD scores at any of the positions.\n",replicates_count);
	else if (!modcalc)
	  sf(ps,"\nNone of the %d replicates exceeded the real data\n"
	        "LOD scores at any of the positions.\n",replicates_count);
	if (modcalc ==2 || !modcalc) {
	  pr(); fpr(pmodfp);
	  sf(ps,"\nThe p-value for all positions is < %lf (increase number\n"
	        "of replicates in order to get a more precise result).\n",
	     (double) 1 / (double) replicates_count);
	  pr(); fpr(pmodfp);
	}
      }
    }
  }

  /* Give distributions for replicate results if desired 
     (only works without sequential simulation option, see above) */
  if (show_distribution && !best_position && modcalc != 1)  {
    if (modcalc)
      sf(ps,"\n\nDistribution of replicate results regarding overall maximum MOD scores\n"
            "in the replicates (best MOD score in real data was %.6lf):\n\n",
            save_max_total_lod);
    else 
      sf(ps,"\n\nDistribution of replicate results regarding overall maximum LOD scores	\n"
            "in the replicates (best LOD score in real data set was %.6lf):\n\n",
            save_max_total_lod);
    pr(); fpr(pmodfp);

    gridpoint_pmod = 0.0;
    exceed_cummulus = 0;

    for (i=0; i<101; i++) {
     exceed_cummulus = exceed_cummulus + exceed_max_total_lod_grid[i]; 
    }

    /* report grid for overall highest scores in replicates */
    for (i=0; i<100; i++) {
      if (i!=0) 
        exceed_cummulus=exceed_cummulus-exceed_max_total_lod_grid[i-1];
      if (exceed_cummulus == 0) {
        early_exit_rv = TRUE;
        break;
      } else { 
        sf(ps,"[%1.2lf / %1.2lf[: %8.1d (%8.1d)  Lower boundary p-value: %7.6lf \n", 
	      gridpoint_pmod, (gridpoint_pmod + 0.10000000000), 
	      exceed_max_total_lod_grid[i], exceed_cummulus, 
              (((double) exceed_cummulus / (double) replicates_count)));
        pr(); fpr(pmodfp);
      } 
      gridpoint_pmod += 0.1;
      if (early_exit_rv)
        break;
    }
    if (!early_exit_rv)
      exceed_cummulus = exceed_cummulus - exceed_max_total_lod_grid[i-1];
    if (exceed_cummulus == 0) {
      sf(ps,"greater %1.2lf : %8.1d\n", 
            gridpoint_pmod, exceed_max_total_lod_grid[100]);
      pr(); fpr(pmodfp); 
    } else {
      sf(ps,"greater %1.2lf : %8.1d  p-value: %7.6lf \n", 
            gridpoint_pmod, exceed_max_total_lod_grid[100], 
            ((double)exceed_max_total_lod_grid[100]/(double)replicates_count));
      pr(); fpr(pmodfp); 
    }

    early_exit_rv = FALSE;

    /* Show the following only for modcalc == 2 or !modcalc */
    if (modcalc == 2 || !modcalc) {
      /* report grid for position with highest score in real data */
      sf(ps,"\n\nDistribution of replicate results regarding position (%.2lf cM)\n"
            "with overall real data maximum score:\n\n", wpos[save_max_pos]);
      pr(); fpr(pmodfp);

      exceed_cummulus = 0;
      for (i=0; i<101; i++) {
        exceed_cummulus = exceed_cummulus + exceed_total_lod_grid[save_max_pos][i];
      }
      gridpoint_pmod = 0.0;
      for (i=0; i<100; i++) {
        if (i!=0) 
          exceed_cummulus=exceed_cummulus-exceed_total_lod_grid[save_max_pos][i-1];
        if (exceed_cummulus == 0) {
          early_exit_rv = TRUE;
          break;
        } else {
          sf(ps,"[%1.2lf / %1.2lf[: %8.1d (%8.1d)  Lower boundary p-value: %7.6lf\n",
                gridpoint_pmod, (gridpoint_pmod+0.1), 
                exceed_total_lod_grid[save_max_pos][i], exceed_cummulus, 
              ((double) exceed_cummulus / (double) replicates_count)); 
          pr(); fpr(pmodfp); 
        }
        gridpoint_pmod += 0.1;
        if (early_exit_rv)
          break;
      }
      if (!early_exit_rv)
        exceed_cummulus = exceed_cummulus - exceed_total_lod_grid[save_max_pos][i-2];
      if (exceed_cummulus == 0) {
        sf(ps,"greater %1.2lf : %8.1d\n", 
             gridpoint_pmod, exceed_total_lod_grid[save_max_pos][100]); 
        pr(); fpr(pmodfp); 
      } else {
        sf(ps,"greater %1.2lf : %8.1d  p-value: %7.6lf\n", 
              gridpoint_pmod, exceed_total_lod_grid[save_max_pos][100], 
            ((double)exceed_total_lod_grid[save_max_pos][100]/(double)replicates_count));
        pr(); fpr(pmodfp); 
      }
      early_exit_rv = FALSE;

      /* report grid for all positions (only calculated positions) */
      sf(ps,"\n\nDistribution of replicate results for all positions:\n");
      pr(); fpr(pmodfp);

      for (j=first_mod_pos; j<=last_mod_pos; j++) {
        sf(ps,"\n position %7.3lf cM: \n", wpos[j]); 
        pr(); fpr(pmodfp);
        exceed_cummulus = 0;
        for (i=0; i<101; i++) {
          exceed_cummulus = exceed_cummulus + exceed_total_lod_grid[j][i];
        }
        gridpoint_pmod = 0.0;
        for (i=0; i<100; i++) {
          if (i!=0) 
            exceed_cummulus=exceed_cummulus-exceed_total_lod_grid[j][i-1];
          if (exceed_cummulus == 0) {
            early_exit_rv = TRUE;
            break;
          } else {
            sf(ps,"[%1.2lf / %1.2lf[: %8.1d (%8.1d)  Lower boundary p-value: %7.6lf\n",
                  gridpoint_pmod, (gridpoint_pmod+0.1), 
                  exceed_total_lod_grid[j][i], exceed_cummulus, 
                ((double) exceed_cummulus / (double) replicates_count)); 
            pr(); fpr(pmodfp); 
          }
          gridpoint_pmod += 0.1;
          if (early_exit_rv)
            break;
        }
        if (!early_exit_rv)
          exceed_cummulus = exceed_cummulus - exceed_total_lod_grid[j][i-1];
        if (exceed_cummulus == 0) {
          sf(ps,"greater %1.2lf : %8.1d\n", 
                gridpoint_pmod, exceed_total_lod_grid[j][100]); 
          pr(); fpr(pmodfp); 
        } else {
          sf(ps,"greater %1.2lf : %8.1d  p-value: %7.6lf\n", 
                gridpoint_pmod, exceed_total_lod_grid[j][100], 
              ((double)exceed_total_lod_grid[j][100]/(double)replicates_count));
          pr(); fpr(pmodfp); 
        }
        early_exit_rv = FALSE;
      } /* for (j=0 ... )*/
    }
  } else if (show_distribution && best_position) {
    /* report grid for position with highest score in real data */
    sf(ps,"\n\nDistribution of replicate results regarding position (%.2lf cM)\n"
          "with overall real data maximum score:\n\n", wpos[save_max_pos]);
    pr(); fpr(pmodfp);

    exceed_cummulus = 0;
    for (i=0; i<101; i++) {
      exceed_cummulus = exceed_cummulus + exceed_total_lod_grid[save_max_pos][i];
    }
    gridpoint_pmod = 0.0;
    for (i=0; i<100; i++) {
      if (i!=0) 
        exceed_cummulus=exceed_cummulus-exceed_total_lod_grid[save_max_pos][i-1];
      if (exceed_cummulus == 0) {
        early_exit_rv = TRUE;
        break;
      } else {
        sf(ps,"[%1.2lf / %1.2lf[: %8.1d (%8.1d)  Lower boundary p-value: %7.6lf\n",
              gridpoint_pmod, (gridpoint_pmod+0.1), 
              exceed_total_lod_grid[save_max_pos][i], exceed_cummulus, 
            ((double) exceed_cummulus / (double) replicates_count)); 
        pr(); fpr(pmodfp); 
      }
      gridpoint_pmod += 0.1;
      if (early_exit_rv)
        break;
    }
    if (!early_exit_rv)
      exceed_cummulus = exceed_cummulus - exceed_total_lod_grid[save_max_pos][i-1];
    if (exceed_cummulus == 0) {
      sf(ps,"greater %1.2lf : %8.1d\n", 
           gridpoint_pmod, exceed_total_lod_grid[save_max_pos][100]); 
      pr(); fpr(pmodfp); 
    } else {
      sf(ps,"greater %1.2lf : %8.1d  p-value: %7.6lf\n", 
            gridpoint_pmod, exceed_total_lod_grid[save_max_pos][100], 
          ((double)exceed_total_lod_grid[save_max_pos][100]/(double)replicates_count));
      pr(); fpr(pmodfp); 
    }
    early_exit_rv = FALSE;
  } else if (show_distribution && modcalc == 1) {
    sf(ps,"\n\nDistribution of replicate results regarding overall maximum MOD scores\n"
          "in replicates (best MOD score in real data was %.6lf):\n\n",
          save_max_total_lod);
    pr(); fpr(pmodfp);

    gridpoint_pmod = 0.0;
    exceed_cummulus = 0;

    for (i=0; i<101; i++) {
     exceed_cummulus = exceed_cummulus + exceed_max_total_lod_grid[i]; 
    }

    /* report grid for overall highest scores in replicates */
    for (i=0; i<100; i++) {
      if (i!=0) 
        exceed_cummulus=exceed_cummulus-exceed_max_total_lod_grid[i-1];
      if (exceed_cummulus == 0) {
        early_exit_rv = TRUE;
        break;
      } else { 
        sf(ps,"[%1.2lf / %1.2lf[: %8.1d (%8.1d)  Lower boundary p-value: %7.6lf \n", 
              gridpoint_pmod, (gridpoint_pmod + 0.10000000000), 
              exceed_max_total_lod_grid[i], exceed_cummulus, 
              (((double) exceed_cummulus / (double) replicates_count)));
        pr(); fpr(pmodfp);
      } 
      gridpoint_pmod += 0.1;
      if (early_exit_rv)
        break;
    }
    if (!early_exit_rv)
      exceed_cummulus = exceed_cummulus - exceed_max_total_lod_grid[i-1];
    if (exceed_cummulus == 0) {
      sf(ps,"greater %1.2lf : %8.1d\n", 
            gridpoint_pmod, exceed_max_total_lod_grid[100]);
      pr(); fpr(pmodfp); 
    } else {
      sf(ps,"greater %1.2lf : %8.1d  p-value: %7.6lf \n", 
            gridpoint_pmod, exceed_max_total_lod_grid[100], 
            ((double)exceed_max_total_lod_grid[100]/(double)replicates_count));
      pr(); fpr(pmodfp); 
    }
    early_exit_rv = FALSE;
  }
}

void check_discarded (int ind)
{
	/* take care of the fact, that with allele-simulation        */
	/* real_num_vecs can change from one replicate to another.   */
	/* (inconsistencies with saved models will appear if         */
	/* family size changes)                                      */

	int i;
	if (!modscore_done && modcalc) {
		for (i=0; i<ind; i++) {
			save_famtree[famnum_pointer[famnum]][i].discard = famtree[famnum][i].discard; 
		}
		diff_discard = FALSE;
	} else if (modcalc) { /* real_num_vecs_calculated skipped because this excludes families that have never been analyzed so far */
		if (saved_model_allocated[famnum_pointer[famnum]] && saved_models) { /* saved_models 'is' TRUE when check_discarded is invoked; addresses pedigrees already analyzed */
			diff_discard = FALSE;
			for (i=0; i<ind; i++) {
				if (save_famtree[famnum_pointer[famnum]][i].discard != famtree[famnum][i].discard) {
					diff_discard = TRUE;
					save_famtree[famnum_pointer[famnum]][i].discard = famtree[famnum][i].discard; 
					if (replicates_store_result && debug_out_pv) {
						sf(ps,"save_famtree[%d][%d].discard: %d, famtree[%d][%d].discard: %d, "
								"diff_discard: %d \n", famnum_pointer[famnum], 
								i, save_famtree[famnum_pointer[famnum]][i].discard, 
								famnum, i, famtree[famnum][i].discard, diff_discard);
						fpr(replifp);
					}
				}
			}
		} else if ((!saved_model_allocated[famnum_pointer[famnum]] && saved_models) || (!saved_models)) { /* !saved_models is impossible!; now addresses pedigrees that are analyzed for the first time */
			for (i=0; i<ind; i++) {
				save_famtree[famnum_pointer[famnum]][i].discard = famtree[famnum][i].discard; 
			}
			diff_discard = FALSE;
		}
	}
}


void check_zero_marker (int ind)
{
  /* NEW for PMOD:************************************************/
  /* check_zero_marker checks for untyped markers (important for */
  /* !simulate_untyped).                                         */
  /***************************************************************/

  int i, j, sum, save_sum, pos;

  for (j=0; j<ind; j++) {
    /* Need this to keep track of untyped markers (!simulate_untyped). */
    array(save_famtree[famnum_count][j].zero_marker, num_markers, int);
    sum=0;
    for (i=0; i<num_in_map_order; i++) {
      save_sum = sum;
      save_famtree[famnum_count][j].zero_marker[map_order[i]] = 0;
      pos = (map_order[i]*2) + allele_offset;
      if (allele_data[j][pos] > 0) sum++;
      if (allele_data[j][pos+1] > 0) sum++;
      if (save_sum == sum) {
        save_famtree[famnum_count][j].zero_marker[map_order[i]] = 1;
      }
    }
    if (sum > 0) {
      save_famtree[famnum_count][j].is_genotyped = TRUE;
    } else {
      save_famtree[famnum_count][j].is_genotyped = FALSE;
    }
  }
}
